# utils/segment_utils.py
import evfuncs
import matplotlib
import numpy as np
import os
import threading
import tkinter as tk
import tkinter.font as tkFont
import torch
import re
from tkinter import ttk, messagebox

matplotlib.use('Agg')


def load_segmentation_checkmarks(all_files):
    """Check whether files have been manually checked as being segmented"""
    unsegment_files = []
    for file in all_files:
        recfile_path = os.path.splitext(file)[0] + ".rec"
        with open(recfile_path, "r") as f:
            content = f.read()

        hand_segmented_pattern = r"Hand Segmented = (\d+)"
        hand_segmented_match = re.search(hand_segmented_pattern, content)

        if hand_segmented_match.group(1) == '0':
            unsegment_files.append(file)

    return unsegment_files


def start_segment_evfuncs(app_state, selection, batch_file, bird_combobox, experiment_combobox, day_combobox):
    """Start the segmentation process using Evfuncs in a separate thread."""
    from moove.utils import get_files_for_day, get_files_for_experiment, get_files_for_bird, get_file_data_by_index

    # Retrieve file selection
    bird, experiment, day = bird_combobox.get(), experiment_combobox.get(), day_combobox.get()
    files = []
    if selection == "current_day":
        files = get_files_for_day(app_state, bird, experiment, day, batch_file)
    elif selection == "current_experiment":
        files = get_files_for_experiment(app_state, bird, experiment, batch_file)
    elif selection == "current_bird":
        files = get_files_for_bird(app_state, bird, batch_file)
    elif selection == "current_file":
        files = [get_file_data_by_index(app_state.data_dir, app_state.song_files, app_state.current_file_index, app_state)["file_path"]]

    # Initialize progress bar
    max_value = len(files)
    progressbar = ttk.Progressbar(app_state.resegment_window, orient=tk.HORIZONTAL, length=200, mode='determinate', maximum=max_value)
    progressbar.grid(row=22, column=0, columnspan=2, pady=(10, 0), sticky="ew")
    
    # Create wrapper function for thread management
    def thread_wrapper():
        current_thread = threading.current_thread()
        try:
            segment_evfuncs(app_state, progressbar, files)
        finally:
            app_state.remove_thread(current_thread)
    
    # Start the segmentation thread
    thread = threading.Thread(target=thread_wrapper, name="SegmentEvfuncsThread")
    app_state.add_thread(thread)
    thread.start()


def segment_evfuncs(app_state, progressbar, files):
    """Perform segmentation using Evfuncs for each selected file."""
    from moove.utils import get_display_data, plot_data, save_notmat, decibel

    # Store the complete original file context to restore it later
    original_data_dir = app_state.data_dir
    original_song_files = app_state.song_files.copy() if app_state.song_files else []
    original_current_file_index = app_state.current_file_index

    for i, file_i in enumerate(files):
        try:
            progressbar['value'] = i
            app_state.resegment_window.update_idletasks()
            
            file_data = get_display_data({"file_name": os.path.basename(file_i), "file_path": file_i}, app_state.config)
            app_state.data_dir = os.path.dirname(file_i)

            # Retrieve segmentation parameters from app_state
            params = app_state.evfuncs_params
            threshold, min_syl_dur = float(params['threshold'].get()), float(params['min_syl_dur'].get())
            min_silent_dur, freq_cutoffs = float(params['min_silent_dur'].get()), tuple(map(int, params['freq_cutoffs'].get().split(',')))
            smooth_window = int(params['smooth_window'].get())

            sampling_rate, rawsong = int(file_data["sampling_rate"]), file_data["song_data"]
            smooth = evfuncs.smooth_data(rawsong, sampling_rate, freq_cutoffs, smooth_window)
            db_smooth = decibel(smooth)
            onsets, offsets = evfuncs.segment_song(db_smooth, sampling_rate, threshold, min_syl_dur, min_silent_dur)

            if onsets is not None and offsets is not None:
                onsets, offsets = np.multiply(onsets, 1000), np.multiply(offsets, 1000)
                file_data.update({"onsets": onsets, "offsets": offsets, "labels": "x" * len(onsets)})
            else:
                file_data.update({"onsets": np.array([]), "offsets": np.array([]), "labels": ""})

            save_notmat(os.path.join(app_state.data_dir, file_data["file_name"] + ".not.mat"), file_data)
            
        except Exception as e:
            app_state.logger.error(f"File {file_i} could not be processed correctly: {e}. Check manually.")
            return

    # Restore the complete original file context
    app_state.data_dir = original_data_dir
    app_state.song_files = original_song_files
    app_state.current_file_index = original_current_file_index

    progressbar['value'] = len(files)
    plot_data(app_state)
    progressbar.grid_forget()
    
    # Schedule Tkinter operations in the main thread
    def show_message():
        messagebox.showinfo("Info", f"Segmentation with Evfuncs completed successfully!")
    
    app_state.resegment_window.after(0, show_message)


def segment_ml(
    model, metadata, device, raw_song_data, sampling_rate=44100,
    chunk_size=64, decision_threshold=0.5, hist_size=3, onset_window_size=5,
    n_onset_true=3, offset_window_size=5, n_offset_false=4, min_silent_duration=0.03,
    min_syllable_length=0.005):
    """Perform segmentation using a machine learning model."""
    from moove.utils.audio_utils import index_to_seconds, seconds_to_chunk_index

    mean_loaded, std_loaded = torch.tensor(metadata['mean']).to(device), torch.tensor(metadata['std']).to(device)
    seg_input_size = hist_size # number of audio chunks used for classification

    y_pred_list, onset_flag = [], False
    onset_idxs, offset_idxs, raw_audio_data_list = [], [], []

    for i in range(0, len(raw_song_data), chunk_size):
        # audio data is processed in chunks
        raw_audio_data = raw_song_data[i:i + chunk_size]
        raw_audio_data_list.append(raw_audio_data)

        # stop if not enough chunks available
        if len(raw_audio_data) < chunk_size:
            break

        if i >= hist_size * chunk_size:
            X = torch.tensor(np.concatenate(raw_audio_data_list[-seg_input_size:]).astype(np.float32)).unsqueeze(0).to(device)
            X = (X - mean_loaded) / std_loaded
            
            y_pred = torch.sigmoid(model(X)).item() # probability of the chunk being part of a syllable
            y_pred_list.append(1 if y_pred > decision_threshold else 0) # if prob > threshold becomes a 1

            sub_y_onset, sub_y_offset = y_pred_list[-onset_window_size:], y_pred_list[-offset_window_size:]
            # if enough 1's in sliding window detected, onset detected
            if not onset_flag and sub_y_onset.count(1) >= n_onset_true:
                # ensure silence was present before detected onset
                if 0 in sub_y_onset[::-1]:
                    onset_idxs.append(len(y_pred_list) - sub_y_onset[::-1].index(0) - 1)
                    onset_flag = True
                else:
                    pass # no valid onset - pass
            # if enough 0's in sliding window detected after onset, offset detected
            elif onset_flag and sub_y_offset.count(0) >= n_offset_false:
                # ensure syllable was present before detected offset
                if 1 in sub_y_offset[::-1]:
                    offset_idxs.append(len(y_pred_list) - sub_y_offset[::-1].index(1) - 1)
                    onset_flag = False
                else:
                    pass # no valid offset - pass

    # Filter segments based on duration thresholds
    segments_2_remove = []
    min_dist, min_dist2 = seconds_to_chunk_index(min_syllable_length, chunk_size, sampling_rate), seconds_to_chunk_index(min_silent_duration, chunk_size, sampling_rate)
    
    if len(onset_idxs) == len(offset_idxs) + 1:
        onset_idxs.pop()
    elif len(onset_idxs) != len(offset_idxs):
        raise ValueError("Mismatch in onset and offset counts; cannot proceed.")

    for i, (onset_idx, offset_idx) in enumerate(zip(onset_idxs, offset_idxs)):
        if offset_idx - onset_idx < min_dist or (i < len(onset_idxs) - 1 and onset_idxs[i + 1] - offset_idx < min_dist2):
            segments_2_remove.append(i)

    # Filter valid onsets and offsets based on duration checks
    onset_idxs = [onset_idxs[i] for i in range(len(onset_idxs)) if i not in segments_2_remove]
    offset_idxs = [offset_idxs[i] for i in range(len(offset_idxs)) if i not in segments_2_remove]

    onsets = [index_to_seconds(idx, chunk_size, sampling_rate) for idx in onset_idxs]
    offsets = [index_to_seconds(idx, chunk_size, sampling_rate) for idx in offset_idxs]

    return onsets, offsets


def create_segmentation_training_dataset(app_state, progressbar, dataset_name, all_files, root):
    """Create a segmentation training dataset from multiple files."""
    from moove.utils import get_display_data, save_features, plot_data, extract_raw_audio

    if len(all_files) == 0:
        messagebox.showinfo("Error", "Not enough files given! You need at least 1 file to create a dataset.")
        return

    chunk_size = int(app_state.train_segmentation_params['chunk_size'].get())
    hist_size = int(app_state.train_segmentation_params['hist_size'].get()) + 1  # Increment for overlapping chunks
    overlap_chunks = app_state.train_segmentation_params['overlap_chunks'].get()

    all_features = []  # List to collect all feature rows

    def generate_concatenated_chunks_with_labels(arr, hist_size, overlap_chunks=False):
        """Concatenate chunks with labels based on history size and overlap setting."""
        concatenated_chunks = []
        step_size = 1 if overlap_chunks else hist_size

        for i in range(0, len(arr) - (hist_size - 1), step_size):
            if np.all(arr[i:i + hist_size, 0] == arr[i, 0]):  # Ensure all rows are from the same file
                file_index = arr[i, 0]  # Preserve file index
                chunk_data = np.concatenate(arr[i:i + hist_size, 1:-1]).flatten()  # Concatenate chunk data
                label = arr[i, -1]  # Preserve label
                concatenated_chunks.append(np.concatenate(([file_index], chunk_data, [label])))

        return np.array(concatenated_chunks)
    
    progressbar.grid_remove()

    # Add running label to the GUI
    font_style = tkFont.Font(family="Arial", size=14)
    running_label = tk.Label(app_state.training_window, text="Looking for segments...", fg="green", font=font_style)
    running_label.grid(row=22, column=0, columnspan=2, pady=(10, 0), sticky=tk.W)
    root.update_idletasks() 

    def get_onset_offset_info(file_path):
        notmat_file = file_path + ".not.mat"
        if os.path.exists(notmat_file):
            notmat_dict = evfuncs.load_notmat(notmat_file)
            return {
                "onsets": notmat_dict.get("onsets", []),
                "offsets": notmat_dict.get("offsets", [])
            }
        else:
            return {"onsets": [], "offsets": []}
        
    num_segs = 0
    for file_path in all_files:
        info = get_onset_offset_info(file_path)
        if len(info["onsets"]) > 0 and len(info["offsets"]) > 0:
            segs = min(len(info["onsets"]), len(info["offsets"]))
            num_segs += segs

    if num_segs == 0:
        running_label.destroy()
        root.update_idletasks() 
        messagebox.showinfo("Error", "No segments found in the given files.")
        return
    
    running_label.destroy()
    root.update_idletasks() 
    progressbar.grid()

    for file_index, file_path in enumerate(all_files):
        display_data = get_display_data({"file_name": os.path.basename(file_path), "file_path": file_path}, app_state.config)
        sampling_rate, rawsong = int(display_data["sampling_rate"]), display_data["song_data"]
        onsets = (np.array(display_data["onsets"]) * sampling_rate / 1000).astype(int)
        offsets = (np.array(display_data["offsets"]) * sampling_rate / 1000).astype(int)
        progressbar['value'] += 1  # Update progress

        # Extract audio features in chunks
        audio_features = extract_raw_audio(rawsong, chunk_size)
        labels = np.zeros(len(audio_features[0]))  # Initialize labels

        # Assign labels based on overlaps with onsets and offsets
        for i, start_idx in enumerate(range(0, len(rawsong) - chunk_size + 1, chunk_size)):
            end_idx = start_idx + chunk_size
            if i < len(labels) and any(onset <= end_idx and start_idx <= offset for onset, offset in zip(onsets, offsets)):
                labels[i] = 1

        # Generate feature rows for this file
        file_features = []
        for i, features in enumerate(np.array(audio_features).T):
            row = np.insert(features, 0, file_index)  # Insert file index as first element
            row = np.append(row, labels[i])  # Append label
            file_features.append(row)

        # Apply concatenation function and extend to all features
        concatenated_features = generate_concatenated_chunks_with_labels(np.array(file_features), hist_size, overlap_chunks)
        if concatenated_features.size > 0:
            all_features.extend(concatenated_features)
    
    # Convert to NumPy array and save
    feature_array = np.array(all_features, dtype=object)
    save_features(app_state, dataset_name, feature_array, chunk_size=chunk_size, hist_size=hist_size, num_syls=num_segs)

    app_state.update_segmentation_datasets_combobox()
    progressbar['value'] = len(all_files)  # Complete progress
    plot_data(app_state)
    progressbar.grid_forget()
    
    # Schedule Tkinter operations in the main thread
    def show_message():
        messagebox.showinfo("Info", "The segmentation training dataset has been created successfully!")
        app_state.change_file(0)
    
    root.after(0, show_message)
    


def segment_files_ml(app_state, progressbar, all_files, model, metadata, device):
    """Segment files using a machine learning model in a threaded process."""
    from moove.utils import get_display_data, plot_data, save_notmat

    # Store the complete original file context to restore it later
    original_data_dir = app_state.data_dir
    original_song_files = app_state.song_files.copy() if app_state.song_files else []
    original_current_file_index = app_state.current_file_index

    # extract parameters used for segmentation
    hist_size, chunk_size = int(metadata['hist_size']), int(metadata['chunk_size'])

    for i, file_path in enumerate(all_files):
        try:
            progressbar['value'] = i  # Update progress
            display_data = get_display_data({"file_name": os.path.basename(file_path), "file_path": file_path}, app_state.config)
            app_state.data_dir = os.path.dirname(file_path)

            sampling_rate = int(display_data["sampling_rate"])
            rawsong = display_data["song_data"]

            # Retrieve segmentation parameters
            params = app_state.mlseg_params
            onsets, offsets = segment_ml(
                model, metadata, device, rawsong, sampling_rate, chunk_size,
                float(params['decision_threshold'].get()), hist_size,
                int(params['onset_window_size'].get()), int(params['n_onset_true'].get()),
                int(params['offset_window_size'].get()), int(params['n_offset_false'].get()),
                float(params['min_silent_duration'].get()), float(params['min_syllable_length'].get())
            )

            # Convert onsets and offsets to milliseconds
            display_data.update({
                "onsets": np.array(onsets) * 1000,
                "offsets": np.array(offsets) * 1000,
                "labels": "x" * len(onsets)  # Label all segments with 'x'
            })
            save_notmat(os.path.join(app_state.data_dir, display_data["file_name"] + ".not.mat"), display_data)
            
        except Exception as e:
            app_state.logger.error(f"File {file_path} could not be processed correctly: {e}. Check manually.")
            return

    # Restore the complete original file context
    app_state.data_dir = original_data_dir
    app_state.song_files = original_song_files
    app_state.current_file_index = original_current_file_index

    # Ensure we have a valid file context for navigation
    if not app_state.song_files or len(app_state.song_files) == 0:
        app_state.logger.debug("No song files found, trying to populate from restored directory")
        try:
            if os.path.exists(app_state.data_dir):
                files_in_dir = [f for f in os.listdir(app_state.data_dir) 
                               if f.lower().endswith('.wav') or f.lower().endswith('.cbin')]
                if files_in_dir:
                    app_state.song_files = files_in_dir
                    app_state.current_file_index = 0
                    app_state.logger.debug(f"Populated song_files with {len(files_in_dir)} files")
                else:
                    app_state.logger.debug("No audio files found in restored directory")
        except Exception as e:
            app_state.logger.debug(f"Could not populate song_files: {e}")

    # Final UI update
    app_state.reset_edit_type()
    plot_data(app_state)
    progressbar.grid_forget()
    
    # Schedule Tkinter operations in the main thread
    def show_message():
        messagebox.showinfo("Info", "Segmentation completed successfully!")
    
    app_state.resegment_window.after(0, show_message)


def start_segment_files_thread(app_state, segmentation_model_name, selection, checkbox_ow, batch_file, bird_combobox, experiment_combobox, day_combobox):
    """Start a threaded process to segment files based on a selected model and criteria."""
    from moove.utils import get_files_for_day, get_files_for_experiment, get_files_for_bird, get_file_data_by_index

    # Retrieve file selection based on user criteria
    bird = bird_combobox.get()
    experiment = experiment_combobox.get()
    day = day_combobox.get()
    files = []

    if selection == "current_day":
        files = get_files_for_day(app_state, bird, experiment, day, batch_file)
    elif selection == "current_experiment":
        files = get_files_for_experiment(app_state, bird, experiment, batch_file)
    elif selection == "current_bird":
        files = get_files_for_bird(app_state, bird, batch_file)
    elif selection == "current_file":
        files = [get_file_data_by_index(app_state.data_dir, app_state.song_files, app_state.current_file_index, app_state)["file_path"]]

    # Set up device and load model with metadata
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    try:
        checkpoint = torch.load(os.path.join(app_state.config['global_dir'], 'trained_models', f'{segmentation_model_name}.pth'), map_location=device)
    except: 
        messagebox.showinfo("Error", "Selected segmentation model doesn't exist or is not valid! Perhaps you forgot to pick a model?")
    model, metadata = checkpoint['model'], checkpoint['metadata']
    
    model.to(device)
    model.eval()
    
    if checkbox_ow:
        files = files
    else:
        # use non-manually segmented files only
        files = load_segmentation_checkmarks(files)

    # Initialize and display progress bar
    max_value = len(files)
    progressbar = ttk.Progressbar(app_state.resegment_window, orient=tk.HORIZONTAL, length=200, mode='determinate', maximum=max_value)
    progressbar.grid(row=22, column=0, columnspan=2, pady=(10, 0), sticky="ew")
    
    # Create wrapper function for thread management
    def thread_wrapper():
        current_thread = threading.current_thread()
        try:
            segment_files_ml(app_state, progressbar, files, model, metadata, device)
        finally:
            app_state.remove_thread(current_thread)
    
    thread = threading.Thread(target=thread_wrapper, name="SegmentMLThread")
    app_state.add_thread(thread)
    thread.start()


def start_create_segmentation_training_dataset(app_state, dataset_name, use_selected_files, selection, batch_file, bird_combobox, experiment_combobox, day_combobox, root):
    """Start a threaded process to create a segmentation training dataset."""
    from moove.utils import get_files_for_day, get_files_for_experiment, get_files_for_bird, filter_segmented_files

    font_style = tkFont.Font(family="Arial", size=14)
    running_label = tk.Label(app_state.training_window, text="Looking for files...", fg="green", font=font_style)
    running_label.grid(row=22, column=0, columnspan=2, pady=(10, 0), sticky=tk.W) 
    root.update_idletasks() 
    
    # Retrieve file selection based on user criteria
    bird = bird_combobox.get()
    experiment = experiment_combobox.get()
    day = day_combobox.get()
    files = []

    if selection == "current_day":
        files = get_files_for_day(app_state, bird, experiment, day, batch_file)
    elif selection == "current_experiment":
        files = get_files_for_experiment(app_state, bird, experiment, batch_file)
    elif selection == "current_bird":
        files = get_files_for_bird(app_state, bird, batch_file)

    if use_selected_files:
        files = filter_segmented_files(files)
        
    running_label.destroy()
    root.update_idletasks() 

    # Initialize and display progress bar
    dataset_name = str(dataset_name)
    # check if dataset name is valid
    if len(dataset_name) < 1:
        messagebox.showinfo("Error", "Dataset name not valid! A dataset name needs to contain at least one character.")
    else:
        # create dataset with valid name
        max_value = len(files)
        progressbar = ttk.Progressbar(app_state.training_window, orient=tk.HORIZONTAL, length=200, mode='determinate', maximum=max_value)
        progressbar.grid(row=22, column=0, columnspan=2, pady=(10, 0), sticky="ew")
        
        # Create wrapper function for thread management
        def thread_wrapper():
            current_thread = threading.current_thread()
            try:
                create_segmentation_training_dataset(app_state, progressbar, dataset_name, files, root)
            finally:
                app_state.remove_thread(current_thread)
        
        thread = threading.Thread(target=thread_wrapper, name="CreateSegDatasetThread")
        app_state.add_thread(thread)
        thread.start()



# utils/clustering_utils.py
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
import threading
import tkinter as tk
import warnings
import evfuncs
import tkinter.font as tkFont
from matplotlib import cm
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
from scipy import interpolate
from scipy.signal import spectrogram
from sklearn.cluster import KMeans
from tkinter import ttk, messagebox
from umap import UMAP

# Disable system warnings
warnings.filterwarnings('ignore')


def start_create_cluster_dataset_thread(app_state, dataset_name, use_selected_files, selection, batch_file, bird_combobox, experiment_combobox, day_combobox, root):
    """Start a thread to create a cluster dataset based on selected files and criteria."""
    from moove.utils import get_files_for_day, get_files_for_experiment, get_files_for_bird, filter_segmented_files
    bird = bird_combobox.get()
    experiment = experiment_combobox.get()
    day = day_combobox.get()

    if selection == "current_day":
        files = get_files_for_day(app_state, bird, experiment, day, batch_file)
    elif selection == "current_experiment":
        files = get_files_for_experiment(app_state, bird, experiment, batch_file)
    elif selection == "current_bird":
        files = get_files_for_bird(app_state, bird, batch_file)

    if use_selected_files:
        files = filter_segmented_files(files)

    app_state.logger.debug(
        "Creating training dataset with parameters: Use selected files: %s, Selection: %s, Batch file: %s", use_selected_files, selection, batch_file
    )

    # check if dataset name is valid
    if len(dataset_name) < 1:
        messagebox.showinfo("Error", "Dataset name not valid! A dataset name needs to contain at least one character.")
    else:
        # create dataset with valid name
        max_value = len(files)
        progressbar = ttk.Progressbar(app_state.cluster_window, orient=tk.HORIZONTAL, length=200, mode='determinate', maximum=max_value)
        progressbar.grid(row=999, column=0, columnspan=2, pady=(10, 0), sticky=tk.W + tk.E)
        
        # Create wrapper function for thread management
        def thread_wrapper():
            current_thread = threading.current_thread()
            try:
                create_cluster_dataset(app_state, dataset_name, progressbar, max_value, files, root)
            finally:
                app_state.remove_thread(current_thread)
        
        thread = threading.Thread(target=thread_wrapper, name="CreateClusterDatasetThread")
        app_state.add_thread(thread)
        thread.start()


def create_cluster_dataset(app_state, dataset_name, progressbar, max_value, all_files, root):
    """Generate and save a cluster dataset, tracking progress with a progress bar."""
    from moove.utils import get_display_data, seconds_to_index, decibel, plot_data
    
    # Store the complete original file context to restore it later
    original_data_dir = app_state.data_dir
    original_song_files = app_state.song_files.copy() if app_state.song_files else []
    original_current_file_index = app_state.current_file_index
    
    if dataset_name:
        going_prod_df = pd.DataFrame(columns=['file', 'onset_no', 'cluster_flattend_spectrogram', 'label'])
        entry_no = 0

    progressbar.grid_remove()

    # Add running label to the GUI
    font_style = tkFont.Font(family="Arial", size=14)
    running_label = tk.Label(app_state.cluster_window, text="Looking for segments...", fg="green", font=font_style)
    running_label.grid(row=22, column=0, columnspan=2, pady=(10, 0), sticky=tk.W)
    root.update_idletasks() 

    def get_onset_offset_info(file_path):
        notmat_file = file_path + ".not.mat"
        if os.path.exists(notmat_file):
            notmat_dict = evfuncs.load_notmat(notmat_file)
            return {
                "onsets": notmat_dict.get("onsets", []),
                "offsets": notmat_dict.get("offsets", [])
            }
        else:
            return {"onsets": [], "offsets": []}
    
    # Count number of segments given for dataset    
    num_segs = 0
    for file_path in all_files:
        info = get_onset_offset_info(file_path)
        if len(info["onsets"]) > 0 and len(info["offsets"]) > 0:
            segs = min(len(info["onsets"]), len(info["offsets"]))
            num_segs += segs

    if num_segs < 10:
        running_label.destroy()
        root.update_idletasks() 
        messagebox.showinfo("Error", "Not enough segments given. Need at least 10 segments to form clusters.")
        return
    
    running_label.destroy()
    root.update_idletasks() 
    progressbar.grid()

    for i in range(max_value):
        progressbar['value'] = i
        file_i = all_files[i]
        file_path = {"file_name": os.path.basename(file_i), "file_path": os.path.join(os.getcwd(), file_i)}
        display_dict = get_display_data(file_path, app_state.config)
        app_state.data_dir = os.path.dirname(file_i)

        sampling_rate = int(display_dict["sampling_rate"])
        rawsong = display_dict["song_data"]

        freq_cutoffs = tuple(map(int, app_state.spec_params['freq_cutoffs'].get().split(',')))
        onsets, offsets = display_dict["onsets"], display_dict["offsets"]

        if dataset_name:
            nperseg = int(app_state.spec_params['nperseg'].get())
            noverlap = int(app_state.spec_params['noverlap'].get())
            nfft = int(app_state.spec_params['nfft'].get())
            
            for syllable_no, (onset, offset) in enumerate(zip(onsets, offsets)):
                entry_no += 1
                onset_index = int(seconds_to_index(onset, sampling_rate))
                offset_index = int(seconds_to_index(offset, sampling_rate))
                cutted_raw_song = rawsong[onset_index:offset_index]
                f, t, Sxx_cluster = spectrogram(cutted_raw_song, fs=sampling_rate, nperseg=nperseg, noverlap=noverlap, nfft=nfft)

                # Filter frequencies within defined range
                Sxx_cluster = Sxx_cluster[(f >= freq_cutoffs[0]) & (f <= freq_cutoffs[1]), :]

                # Linear interpolation
                original_shape = Sxx_cluster.shape
                x_old = np.linspace(0, 1, original_shape[1])
                x_new = np.linspace(0, 1, 40)
                f_i = interpolate.interp1d(x_old, Sxx_cluster, kind='linear', axis=1)
                Sxx_cluster = f_i(x_new)
                Sxx_cluster = decibel(Sxx_cluster)

                going_prod_df.loc[entry_no] = [file_i, syllable_no, Sxx_cluster.flatten(), "x"]

    if dataset_name:
        file_path = os.path.join(app_state.config['global_dir'], 'cluster_data', f'{dataset_name}_clus.pkl')
        going_prod_df.to_pickle(file_path)
        app_state.update_cluster_datasets_combobox()

    # Restore the complete original file context
    app_state.data_dir = original_data_dir
    app_state.song_files = original_song_files
    app_state.current_file_index = original_current_file_index

    progressbar['value'] = max_value
    progressbar.grid_forget()
    
    # Schedule Tkinter operations in the main thread
    def show_message():
        messagebox.showinfo("Info", f"Cluster dataset '{dataset_name}' created successfully!")
    
    root.after(0, show_message)


def start_clustering_thread(root, app_state, dataset_name_entry):
    """Start the clustering process in a separate thread."""
    dataset_name = dataset_name_entry
    if dataset_name == "Select Cluster Dataset":
        messagebox.showinfo("Error", "Selected cluster dataset not valid! Perhaps you forgot to pick a dataset?")
    else:
        messagebox.showinfo("Info", "Clustering started. This may take a while, please wait!")
    
    # Create wrapper function for thread management
    def thread_wrapper():
        current_thread = threading.current_thread()
        try:
            run_clustering(root, app_state, dataset_name)
        finally:
            app_state.remove_thread(current_thread)
    
    thread = threading.Thread(target=thread_wrapper, name="ClusteringThread")
    app_state.add_thread(thread)
    thread.start()


def run_clustering(root, app_state, dataset_name):
    """Run the clustering process using UMAP and KMeans."""
    dataset_name = f"{dataset_name}.pkl"
    n_syllables = int(app_state.umap_k_means_params['n_clusters'].get())
    n_neighbors = int(app_state.umap_k_means_params['n_neighbors'].get())
    min_dist = float(app_state.umap_k_means_params['min_dist'].get())

    font_style = tkFont.Font(family="Arial", size=14)
    running_label = tk.Label(app_state.cluster_window, text="Running...", fg="green", font=font_style)
    running_label.grid(row=999, column=0, columnspan=2, pady=(10, 0), sticky=tk.W)
    root.update_idletasks()

    # Load the dataset 
    dataset_path = os.path.join(os.path.join(app_state.config['global_dir'], 'cluster_data', f'{dataset_name}'))
    if not os.path.exists(dataset_path):
        app_state.logger.error("Dataset %s not found in cluster_data folder.", dataset_name)
        return

    df = pd.read_pickle(dataset_path)
    spectrogram_feature_array = np.array([np.array(x) for x in df['cluster_flattend_spectrogram'].values])

    # Run UMAP
    umap_model = UMAP(n_neighbors=n_neighbors, min_dist=min_dist, n_components=2, metric='euclidean', random_state=42)
    low_dimensional_data = umap_model.fit_transform(spectrogram_feature_array)

    # KMeans Clustering
    kmeans = KMeans(n_clusters=n_syllables, random_state=42)
    labels = kmeans.fit_predict(low_dimensional_data)

    # Map numerical labels to alphabetic characters
    label_mapping = {i: chr(97 + i) for i in range(n_syllables)}
    alphabet_labels = [label_mapping[label] for label in labels]
    df['clustered_label'] = alphabet_labels

    # Add UMAP coordinates to the DataFrame
    df['UMAP1'] = low_dimensional_data[:, 0]
    df['UMAP2'] = low_dimensional_data[:, 1]

    # Save the clustered data
    output_path = os.path.join(app_state.config['global_dir'], 'cluster_data', dataset_name)
    df.to_pickle(output_path)

    # Plot and save the clusters
    plot_clusters(root, app_state, low_dimensional_data, alphabet_labels, output_path)

    running_label.destroy()

    app_state.logger.debug("Clustering complete. Results saved to %s", output_path)
    
    # Schedule Tkinter operations in the main thread
    def show_message():
        messagebox.showinfo("Info", f"Clustering complete! Results saved to {output_path}")
    
    root.after(0, show_message)


def plot_clusters(root, app_state, low_dimensional_data, labels, output_path):
    """Plot and save the clustering results in a new window."""

    def plot():
        unique_labels = sorted(set(labels))
        label_mapping = {label: idx for idx, label in enumerate(unique_labels)}
        numeric_labels = [label_mapping[label] for label in labels]

        # Create a new window for the plot
        plot_window = tk.Toplevel(root)
        plot_window.title("Cluster Plot")

        fig, ax = plt.subplots(figsize=(10, 8))
        scatter = ax.scatter(low_dimensional_data[:, 0], low_dimensional_data[:, 1], 
                             c=numeric_labels, s=5, cmap=cm.get_cmap('jet'))

        # Create a custom legend with alphabetic labels
        handles, _ = scatter.legend_elements()
        legend = ax.legend(handles, unique_labels, title="Labels")

        ax.add_artist(legend)
        ax.set_title("UMAP Clustering")
        ax.set_xlabel("UMAP1")
        ax.set_ylabel("UMAP2")

        # Save the plot
        plot_path = output_path.replace('_clus.pkl', '_clusters.png')
        plt.savefig(plot_path)
        app_state.logger.debug("Cluster plot saved to %s", plot_path)

        # Embed the plot in the new window
        canvas = FigureCanvasTkAgg(fig, master=plot_window)
        canvas.draw()
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)
        toolbar = NavigationToolbar2Tk(canvas, plot_window)
        toolbar.update()
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

    root.after(0, plot)


def replace_labels_from_df(app_state, dataset_name, root=None):
    """Replace labels in the dataset based on clustering results."""
    from moove.utils.file_utils import get_display_data
    from moove.utils.movefuncs_utils import save_notmat
    from moove.utils.plot_utils import plot_data

    if dataset_name == "Select Cluster Dataset":
        messagebox.showinfo("Error", "Selected cluster dataset not valid! Perhaps you forgot to pick a dataset?")
    else:
        messagebox.showinfo("Info", "Replacement of syllables started. This may take a while, please wait!")

    # Store the complete original file context to restore it later
    original_data_dir = app_state.data_dir
    original_song_files = app_state.song_files.copy() if app_state.song_files else []
    original_current_file_index = app_state.current_file_index

    dataset_path = os.path.join(app_state.config['global_dir'], 'cluster_data', f'{dataset_name}.pkl')
    df = pd.read_pickle(dataset_path)
    files = df['file'].unique()

    app_state.logger.debug("Starting replacement of syllables with dataset %s", dataset_name)

    max_value = len(files)
    progressbar = ttk.Progressbar(app_state.cluster_window, orient=tk.HORIZONTAL, length=200, mode='determinate', maximum=max_value)
    progressbar.grid(row=999, column=0, columnspan=2, pady=(10, 0), sticky=tk.W + tk.E)

    for i, file in enumerate(files):
        try:
            progressbar['value'] = i
            progressbar.update()
            # Concatenate all the clustered labels for the current file into a single string
            # Only use clustered_label (from clustering/Dash), not the original 'label' column
            if 'clustered_label' not in df.columns:
                raise KeyError("Dataset has not been clustered yet. Please cluster the dataset first before replacing labels.")
            labels = df.loc[df['file'] == file]['clustered_label'].astype(str).str.cat(sep='')

            # Get display data for the current file
            display_dict = get_display_data({"file_name": os.path.basename(file), "file_path": file}, app_state.config)
            display_dict["labels"] = labels

            app_state.data_dir = os.path.dirname(file)
            save_path = os.path.join(app_state.data_dir, f"{display_dict['file_name']}.not.mat")
            app_state.logger.debug("Saving labels to %s", save_path)

            # Save modified labels to the .not.mat file
            save_notmat(save_path, display_dict)
            
        except Exception as e:
            app_state.logger.error(f"File {file} could not be processed correctly: {e}. Check manually.")
            return

    # Restore the original data_dir so file navigation continues to work
    app_state.data_dir = original_data_dir
    app_state.song_files = original_song_files
    app_state.current_file_index = original_current_file_index
    progressbar['value'] = max_value
    
    # Final UI update
    progressbar.grid_forget()
    app_state.reset_edit_type()
    plot_data(app_state)
    
    # Schedule Tkinter operations in the main thread if root is available
    if root:
        def show_message():
            messagebox.showinfo("Info", "Replacement of syllables complete!")
        root.after(0, show_message)
    else:
        messagebox.showinfo("Info", "Replacement of syllables complete!")





import torch
import torch.nn as nn

class ConvMLP(nn.Module):
    def __init__(self, input_size=256):
        super(ConvMLP, self).__init__()
        
        self.audio_encoder = nn.Sequential(
            nn.Conv1d(1, 16, kernel_size=3, stride=2, padding=1),
            nn.ReLU(),
            nn.Conv1d(16, 32, kernel_size=3, stride=2, padding=1),
            nn.ReLU(),
            nn.Conv1d(32, 64, kernel_size=3, stride=2, padding=1),
            nn.ReLU(),
            nn.Flatten()
        )
        
        with torch.no_grad():
            sample_input = torch.randn(1, 1, input_size)
            self.flat_features = self.audio_encoder(sample_input).shape[1]

        self.mlp_layers = nn.Sequential(
            nn.Linear(self.flat_features, 256),
            nn.BatchNorm1d(256),
            nn.ReLU(),
            nn.Dropout(0.5),
            nn.Linear(256, 128),
            nn.BatchNorm1d(128),
            nn.ReLU(),
            nn.Dropout(0.5),
            nn.Linear(128, 64),
            nn.ReLU(),
            nn.Linear(64, 1)
        )
        
    def forward(self, x):
        x = x.unsqueeze(1)
        x = self.audio_encoder(x)
        x = self.mlp_layers(x)
        return x



# utils/dash_utils.py
import ctypes
import dash
import numpy as np
import pandas as pd
import plotly.graph_objs as go
import socket
import string
import threading
import os
import webbrowser
import logging
from dash import dcc, html
from dash.dependencies import Input, Output, State
from flask import Flask
from wsgiref.simple_server import make_server, WSGIRequestHandler
from tkinter import ttk, messagebox

# Suppress Flask/Werkzeug/Dash logging to reduce console output
logging.getLogger('werkzeug').setLevel(logging.ERROR)
logging.getLogger('flask').setLevel(logging.ERROR)
logging.getLogger('dash').setLevel(logging.ERROR)


class QuietWSGIRequestHandler(WSGIRequestHandler):
    """Custom request handler that suppresses HTTP request logging."""
    def log_message(self, format, *args):
        """Override to suppress all HTTP request logs."""
        pass  # Do nothing - suppress all logs

# Global variables to store labels and point size
labels = None
point_size = 5


def run_flask_server(app_state, dataset_name):
    """Create and run a Flask server for the Dash app."""
    
    server = Flask(__name__)
    app = dash.Dash(__name__, server=server)

    global labels, point_size
    df = pd.read_pickle(os.path.join(app_state.config['global_dir'], 'cluster_data', f'{dataset_name}.pkl'))
    # Use clustered_label (created after clustering/Dash editing)
    if 'clustered_label' not in df.columns:
        raise KeyError("Dataset has not been clustered yet. Please cluster the dataset first.")
    labels = df['clustered_label'].values
    low_dimensional_data = df[['UMAP1', 'UMAP2']].values

    unique_labels = list(string.ascii_lowercase)
    label_mapping = {label: idx for idx, label in enumerate(unique_labels)}
    numeric_labels = [label_mapping[label] for label in labels]

    common_style = {'font-family': 'Arial', 'font-size': '14px', 'height': '15px', 'width': '30px', 'margin': '2px'}
    button_style = {'font-family': 'Arial', 'font-size': '13px', 'height': '20px', 'margin': '2px'}

    app.layout = html.Div([
        dcc.Graph(
            id='scatter-plot',
            figure={
                'data': [
                    go.Scatter(
                        x=low_dimensional_data[:, 0],
                        y=low_dimensional_data[:, 1],
                        mode='markers',
                        marker=dict(color=numeric_labels, colorscale='Jet', size=point_size),
                        selected={'marker': {'color': 'red'}},
                        unselected={'marker': {'opacity': 0.5}},
                        text=labels,
                        hoverinfo='text',
                    )
                ],
                'layout': go.Layout(
                    title=dict(
                        text='UMAP Clustering',
                        font=dict(size=24, family='Arial', color='black')
                    ),
                    clickmode='event+select',
                    margin=dict(l=20, r=20, t=40, b=20),
                    autosize=True
                )
            },
            style={'width': '100%', 'height': '100%', 'flex': '1 1 auto'}
        ),
        html.Div([
            html.Div([
                html.Label("Label for selected points:", style={**common_style, 'width': 'auto'}),
                dcc.Input(id='input-label', type='text', style={**common_style, 'margin-right': '5px'}),
                html.Button('Apply', id='apply-button', style=button_style),
                
                html.Label("Change all labels from:", style={**common_style, 'width': 'auto'}),
                dcc.Input(id='input-label-old', type='text', style={**common_style, 'margin-right': '5px'}),
                html.Label("to:", style={**common_style, 'width': 'auto'}),
                dcc.Input(id='input-label-new', type='text', style={**common_style, 'margin-right': '5px'}),
                html.Button('Change All', id='change-all-button', style=button_style),
            ], style={
                'display': 'flex', 
                'align-items': 'center', 
                'flex-wrap': 'nowrap', 
                'min-width': '0'
            }),
            html.Div([
                html.Button('Save', id='save-button', style=button_style),
            ], style={
                'display': 'flex', 
                'justify-content': 'center', 
                'align-items': 'center', 
                'min-width': '0'
            }),
            html.Div([
                html.Button('Increase Point Size', id='increase-size-button', style=button_style),
                html.Button('Decrease Point Size', id='decrease-size-button', style=button_style),
            ], style={
                'display': 'flex', 
                'align-items': 'center', 
                'justify-content': 'flex-end',
                'flex-wrap': 'nowrap', 
                'min-width': '0'
            }),
        ], style={
            'display': 'flex', 
            'justify-content': 'space-between', 
            'flex-wrap': 'nowrap', 
            'width': '100%',  
            'margin-top': '10px',
            'padding': '0px',
            'align-items': 'center'
        }),
        html.Div(id='output-container')
    ], style={
        'display': 'flex', 
        'flex-direction': 'column', 
        'height': '100vh',
        'padding': '10px', 
        'box-sizing': 'border-box'
    })

    @app.callback(
        Output('scatter-plot', 'figure'),
        [
            Input('change-all-button', 'n_clicks'),
            Input('apply-button', 'n_clicks'),
            Input('increase-size-button', 'n_clicks'),
            Input('decrease-size-button', 'n_clicks')
        ],
        [
            State('input-label-old', 'value'),
            State('input-label-new', 'value'),
            State('scatter-plot', 'figure'),
            State('scatter-plot', 'selectedData'),
            State('input-label', 'value')
        ],
        prevent_initial_call=True
    )
    def update_labels_and_point_size(n_clicks_change_all, n_clicks_apply, n_clicks_increase, n_clicks_decrease,
                                     old_label, new_label, figure, selectedData, input_label):
        """Update the labels of selected points, change all labels, or adjust the point size."""
        global labels, point_size

        if n_clicks_change_all and old_label and new_label:
            labels = np.where(labels == old_label, new_label, labels)

        if n_clicks_apply and selectedData and input_label:
            points_indices = [point['pointIndex'] for point in selectedData['points']]
            for idx in points_indices:
                labels[idx] = input_label

        ctx = dash.callback_context
        if ctx.triggered:
            triggered_id = ctx.triggered[0]['prop_id'].split('.')[0]
            if triggered_id == 'increase-size-button':
                point_size = min(point_size + 1, 20)
            elif triggered_id == 'decrease-size-button':
                point_size = max(point_size - 1, 1)

        numeric_labels = [label_mapping[label] for label in labels]
        trace = figure['data'][0]
        trace['marker']['color'] = numeric_labels
        trace['marker']['size'] = point_size
        hover_text = [f'Label: {label}' for label in labels]
        trace['text'] = hover_text
        return figure

    @app.callback(
        Output('output-container', 'children'),
        Input('save-button', 'n_clicks'),
        prevent_initial_call=True
    )
    def save_labels(n_clicks):
        """Save the updated labels to the dataset file and show popup notification."""
        global labels
        df['clustered_label'] = labels
        df.to_pickle(os.path.join(app_state.config['global_dir'], 'cluster_data', f'{dataset_name}.pkl'))
        
        # Show Tkinter popup message instead of Dash message
        try:
            messagebox.showinfo("Success", f"✅ Labels saved successfully!\n\nFile: {dataset_name}.pkl")
        except:
            pass  # In case GUI is not available
        
        # Return empty div (no browser message)
        return html.Div()

    app_state.logger.debug(f"Starting Flask server for dataset: {dataset_name}")
    app_state.server = make_server('localhost', 8050, server, handler_class=QuietWSGIRequestHandler)
    app_state.server_thread = threading.Thread(target=app_state.server.serve_forever, name="DashServerThread")
    
    # Register the server thread
    app_state.add_thread(app_state.server_thread)
    
    app_state.server_thread.start()
    threading.Timer(1, lambda: webbrowser.open("http://127.0.0.1:8050")).start()


def start_dash_app_thread(app_state, dataset_name):
    """Start the Dash application in a separate thread using AppState."""

    if dataset_name == "Select Cluster Dataset":
        messagebox.showinfo("Error", "Selected cluster dataset not valid! Perhaps you forgot to pick a dataset?")
    else:
        if app_state.dash_thread and app_state.dash_thread.is_alive():
            app_state.logger.debug("Dash app is already running.")
            return

        if not is_port_available(8050):
            app_state.logger.warning("Port 8050 is already in use.")
            return

        # Start the Flask server and Dash app (this already creates a thread)
        run_flask_server(app_state, dataset_name)
        
        # Set the dash_thread to the server_thread (no need for double threading)
        app_state.dash_thread = app_state.server_thread
        app_state.logger.debug("Dash app thread started.")


def is_port_available(port):
    """Check if a specific port is available."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) != 0


def kill_thread(thread):
    """Forcefully terminate the given thread."""
    if thread.is_alive():
        _async_raise(thread.ident, SystemExit)


def _async_raise(tid, exctype):
    """Raise the specified exception type in the context of the specified thread."""
    if not isinstance(tid, int):
        raise TypeError("Thread ID must be an integer")
    if not isinstance(exctype, (type, type(None))) or not issubclass(exctype, BaseException):
        raise TypeError("Exception type must be a subclass of BaseException")
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), ctypes.py_object(exctype))
    if res == 0:
        raise ValueError("Invalid thread ID")
    elif res > 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), None)
        raise SystemError("PyThreadState_SetAsyncExc failed")


def stop_dash_app_thread(app_state):
    """Properly stop the Dash application by shutting down the server and threads."""
    app_state.logger.debug("=== STOP DASH APP CALLED ===")
    
    try:
        # Stop the Flask server first
        if hasattr(app_state, 'server') and app_state.server:
            app_state.logger.debug("Shutting down Flask server.")
            app_state.server.shutdown()
            app_state.server = None
            app_state.logger.debug("Flask server shutdown complete.")
        else:
            app_state.logger.debug("No Flask server found to shutdown.")
        
        # Stop the server thread
        if hasattr(app_state, 'server_thread') and app_state.server_thread and app_state.server_thread.is_alive():
            app_state.logger.debug("Stopping server thread.")
            app_state.server_thread.join(timeout=2)  # Wait up to 2 seconds
            app_state.server_thread = None
            app_state.logger.debug("Server thread stopped.")
        else:
            app_state.logger.debug("No server thread found to stop.")
        
        # Stop the dash thread
        if app_state.dash_thread and app_state.dash_thread.is_alive():
            app_state.logger.debug("Stopping Dash application thread.")
            try:
                # Try graceful shutdown first
                app_state.dash_thread.join(timeout=2)  # Wait up to 2 seconds
                app_state.logger.debug("Dash thread joined gracefully.")
            except Exception as join_error:
                # If graceful shutdown fails, force kill
                app_state.logger.debug(f"Graceful join failed: {join_error}. Forcefully terminating the Dash server thread.")
                kill_thread(app_state.dash_thread)
                app_state.logger.debug("Dash thread force killed.")
            
            app_state.dash_thread = None
        else:
            app_state.logger.debug("No Dash thread found to stop.")
        
        app_state.logger.debug("=== DASH APP SUCCESSFULLY STOPPED ===")
        
        # Show confirmation message
        try:
            messagebox.showinfo("Dash GUI", "Dash GUI closed successfully!")
        except:
            pass
        
    except Exception as e:
        app_state.logger.error(f"Error while stopping Dash app: {e}")
        # Fallback: force kill if everything else fails
        if app_state.dash_thread and app_state.dash_thread.is_alive():
            try:
                kill_thread(app_state.dash_thread)
                app_state.dash_thread = None
                app_state.logger.debug("Fallback force kill successful.")
            except Exception as kill_error:
                app_state.logger.error(f"Failed to force kill thread: {kill_error}")



# utils/window_utils.py
import os
import tkinter as tk
from tkinter import ttk


def open_resegment_window(root, app_state, bird_combobox, experiment_combobox, day_combobox):
    """Open the resegmentation window for segmenting the current file using Evfuncs or the Segmentation Network."""
    from moove.utils import start_segment_evfuncs, start_segment_files_thread

    resegment_window = tk.Toplevel(root)
    resegment_window.title("Resegmentation")
    app_state.resegment_window = resegment_window

    resegment_window.geometry("400x450")
    resegment_window.grid_rowconfigure(0, weight=1)
    resegment_window.grid_columnconfigure(0, weight=1)
    resegment_window.grid_columnconfigure(1, weight=1)
    # ensure the window stays in front
    resegment_window.transient(root)

    container_frame = tk.Frame(resegment_window)
    container_frame.grid(row=0, column=0, padx=10, pady=10, sticky="nsew")
    container_frame.grid_columnconfigure(0, weight=1)
    container_frame.grid_columnconfigure(1, weight=1)

    # Left frame for Evfuncs
    left_frame = tk.Frame(container_frame)
    left_frame.grid(row=0, column=0, padx=10, pady=10, sticky="nsew")
    left_frame.grid_columnconfigure(0, weight=1)
    left_frame.grid_columnconfigure(1, weight=1)

    row = 0
    tk.Label(left_frame, text="Evfuncs", font=("Arial", 16)).grid(row=row, column=0, columnspan=2, pady=10, sticky="ew")
    row += 1

    def update_batch_combobox_resegment_ev():
        app_state.update_batch_select_combobox_resegment_ev(select_path=ev_selection_var.get())

    # Radio buttons for Evfuncs selection
    ev_selection_var = tk.StringVar(value="current_file")
    tk.Radiobutton(left_frame, text="Current File", variable=ev_selection_var, value="current_file").grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(left_frame, text="Current Day", variable=ev_selection_var, value="current_day",
                   command=update_batch_combobox_resegment_ev).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(left_frame, text="Current Experiment", variable=ev_selection_var, value="current_experiment",
                   command=update_batch_combobox_resegment_ev).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(left_frame, text="Current Bird", variable=ev_selection_var, value="current_bird",
                   command=update_batch_combobox_resegment_ev).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1

    # Combobox for batch file selection
    ev_batch_file_var = tk.StringVar()
    resegment_batch_combobox_ev = ttk.Combobox(left_frame,textvariable=ev_batch_file_var)
    resegment_batch_combobox_ev.set("Select Batch File")
    app_state.resegment_window.resegment_batch_combobox_ev = resegment_batch_combobox_ev
    app_state.update_batch_select_combobox_resegment_ev(select_path = ev_selection_var.get())
    resegment_batch_combobox_ev.grid(row=2, column=1, sticky="ew")

    evfuncs_params = [
        ("Threshold:", 'threshold'),
        ("Min Syllable Length:", 'min_syl_dur'),
        ("Min Silent Duration:", 'min_silent_dur'),
        ("Frequency Cutoffs:", 'freq_cutoffs'),
        ("Smoothing Window:", 'smooth_window')
    ]

    for label_text, param_key in evfuncs_params:
        tk.Label(left_frame, text=label_text).grid(row=row, column=0, sticky="w")
        entry = tk.Entry(left_frame, textvariable=app_state.evfuncs_params[param_key])
        entry.grid(row=row, column=1, sticky="ew")
        row += 1

    segment_btn_evfuncs = tk.Button(
        left_frame, text="Segment",
        command=lambda: start_segment_evfuncs(app_state, ev_selection_var.get(), ev_batch_file_var.get(), bird_combobox, experiment_combobox, day_combobox)
    )
    segment_btn_evfuncs.grid(row=row, column=0, columnspan=2, pady=(10, 0), sticky="ew")

    # Right frame for Segmentation Network
    right_frame = tk.Frame(container_frame)
    right_frame.grid(row=0, column=1, padx=10, pady=10, sticky="nsew")
    right_frame.grid_columnconfigure(0, weight=1)
    right_frame.grid_columnconfigure(1, weight=1)

    row = 0
    tk.Label(right_frame, text="Segmentation Network", font=("Arial", 16)).grid(row=row, column=0, columnspan=2, pady=10, sticky="ew")
    row += 1

    def update_batch_combobox_resegment():
        app_state.update_batch_select_combobox_resegment(select_path=sm_selection_var.get())

    # Radio buttons for Segmentation Network selection
    sm_selection_var = tk.StringVar(value="current_file")
    tk.Radiobutton(right_frame, text="Current File", variable=sm_selection_var, value="current_file").grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(right_frame, text="Current Day", variable=sm_selection_var, value="current_day",
                   command=update_batch_combobox_resegment).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(right_frame, text="Current Experiment", variable=sm_selection_var, value="current_experiment",
                   command=update_batch_combobox_resegment).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(right_frame, text="Current Bird", variable=sm_selection_var, value="current_bird",
                   command=update_batch_combobox_resegment).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1

    # Combobox for batch file selection
    batch_file_var = tk.StringVar()
    resegment_batch_combobox = ttk.Combobox(right_frame,textvariable=batch_file_var)
    resegment_batch_combobox.set("Select Batch File")
    app_state.resegment_window.resegment_batch_combobox = resegment_batch_combobox
    app_state.update_batch_select_combobox_resegment(select_path = sm_selection_var.get())
    resegment_batch_combobox.grid(row=2, column=1, sticky="ew")

    checkbox_var = tk.BooleanVar()
    tk.Checkbutton(right_frame, text="Overwrite Already Segmented Files", variable=checkbox_var).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1

    # Model selection combobox
    trained_models_path = os.path.join(app_state.config["global_dir"], "trained_models")
    model_files = [f[:-4] for f in os.listdir(trained_models_path) if f.endswith("_seg_model.pth")]
    app_state.current_segmentation_model = tk.StringVar()
    model_combobox = ttk.Combobox(right_frame, textvariable=app_state.current_segmentation_model)
    model_combobox['values'] = model_files
    model_combobox.set("Select Trained Segmentation Model")
    model_combobox.grid(row=row, column=0, columnspan=2, sticky="ew")
    row += 1

    params = [
        ("Decision Threshold:", 'decision_threshold'),
        ("Onset Window Size:", 'onset_window_size'),
        ("N Onset True:", 'n_onset_true'),
        ("Offset Window Size:", 'offset_window_size'),
        ("N Offset False:", 'n_offset_false'),
        ("Min Syllable Length:", 'min_syllable_length'),
        ("Min Silent Duration:", 'min_silent_duration')
    ]

    for label_text, param_key in params:
        tk.Label(right_frame, text=label_text).grid(row=row, column=0, sticky="w")
        entry = tk.Entry(right_frame, textvariable=app_state.mlseg_params[param_key])
        entry.grid(row=row, column=1, sticky="ew")
        row += 1

        # Additional spacing for readability after certain parameters
        if param_key in ['chunk_size', 'decision_threshold', 'n_offset_false']:
            row += 1

    # Segment button for Segmentation Network
    segment_btn_ml = tk.Button(
        right_frame, text="Segment",
        command=lambda: start_segment_files_thread(app_state, app_state.current_segmentation_model.get(),
                                                    sm_selection_var.get(), checkbox_var.get(), batch_file_var.get(),
                                                      bird_combobox, experiment_combobox, day_combobox)
    )
    segment_btn_ml.grid(row=row, column=0, columnspan=2, pady=(10, 0), sticky="ew")

    # Adjust window size to content
    resegment_window.update_idletasks()
    resegment_window.minsize(resegment_window.winfo_reqwidth(), resegment_window.winfo_reqheight())


def open_relabel_window(root, app_state, bird_combobox, experiment_combobox, day_combobox):
    """Open the relabel window for relabeling the current file using the Classification Network."""
    from moove.utils import start_classify_files_thread

    relabel_window = tk.Toplevel(root)
    relabel_window.title("Relabel")
    app_state.relabel_window = relabel_window
    # ensure the window stays in front
    relabel_window.transient(root)

    relabel_window.geometry("400x280")
    relabel_window.grid_rowconfigure(0, weight=1)
    relabel_window.grid_columnconfigure(0, weight=1)

    # Container Frame for layout
    container_frame = tk.Frame(relabel_window)
    container_frame.grid(row=0, column=0, sticky="nsew", padx=10, pady=10)
    container_frame.grid_rowconfigure(0, weight=1)
    container_frame.grid_columnconfigure(0, weight=1)

    # Frame for content
    content_frame = tk.Frame(container_frame)
    content_frame.grid(row=0, column=0, sticky="nsew")
    content_frame.grid_rowconfigure(99, weight=1)
    content_frame.grid_columnconfigure(0, weight=1)
    content_frame.grid_columnconfigure(1, weight=1)

    row = 0
    # Header Label
    tk.Label(content_frame, text="Classification Network", font=("Arial", 16)).grid(row=row, column=0, columnspan=2, pady=10, sticky="nsew")
    row += 1

    def update_batch_combobox_relabel():
        app_state.update_batch_select_combobox_relabel(select_path=selection_var.get())

    # Radio buttons for selection
    selection_var = tk.StringVar(value="current_file")
    tk.Radiobutton(content_frame, text="Current File", variable=selection_var, value="current_file").grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(content_frame, text="Current Day", variable=selection_var, value="current_day",
                   command=update_batch_combobox_relabel).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(content_frame, text="Current Experiment", variable=selection_var, value="current_experiment",
                   command=update_batch_combobox_relabel).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(content_frame, text="Current Bird", variable=selection_var, value="current_bird",
                   command=update_batch_combobox_relabel).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1

    # Combobox for batch file selection
    batch_file_var = tk.StringVar()
    relabel_batch_combobox = ttk.Combobox(content_frame,textvariable=batch_file_var, width=5)
    relabel_batch_combobox.set("Select Batch File")
    app_state.relabel_window.relabel_batch_combobox = relabel_batch_combobox
    app_state.update_batch_select_combobox_relabel(select_path = selection_var.get())
    relabel_batch_combobox.grid(row=2, column=1, sticky="ew")

    checkbox_var = tk.BooleanVar()
    tk.Checkbutton(content_frame, text="Overwrite Already Classified Files", variable=checkbox_var).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1

    # ComboBox for selecting models
    trained_models_path = os.path.join(app_state.config["global_dir"], "trained_models")
    model_files = [f[:-4] for f in os.listdir(trained_models_path) if f.endswith("_class_model.pth")]
    app_state.current_classification_model = tk.StringVar()
    model_combobox = ttk.Combobox(content_frame, textvariable=app_state.current_classification_model)
    model_combobox['values'] = model_files
    model_combobox.set("Select Trained Classification Model")
    model_combobox.grid(row=row, column=0, columnspan=2, sticky="ew")
    row += 1

    # Relabel Button
    segment_btn_ml = tk.Button(content_frame, text="Relabel", command=lambda: start_classify_files_thread(app_state, app_state.current_classification_model.get(),
                                                                                                           selection_var.get(), checkbox_var.get(), batch_file_var.get(),
                                                                                                             bird_combobox, experiment_combobox, day_combobox))
    segment_btn_ml.grid(row=row, column=0, columnspan=2, sticky="ew")

    # Adjust window size to content
    relabel_window.update_idletasks()
    relabel_window.minsize(relabel_window.winfo_reqwidth(), relabel_window.winfo_reqheight())


def open_training_window(root, app_state, bird_combobox, experiment_combobox, day_combobox):
    """Open the training window for training the Segmentation and Classification Networks."""
    from moove.utils import (
        start_create_segmentation_training_dataset, start_segmentation_training,
        start_create_classification_training_dataset, start_classification_training, 
        find_batch_files
    )

    training_window = tk.Toplevel(root)
    training_window.title("Training")
    training_window.geometry("400x540")
    app_state.training_window = training_window
    # ensure the window stays in front
    training_window.transient(root)

    training_window.grid_rowconfigure(0, weight=1)
    training_window.grid_columnconfigure(0, weight=1)
    training_window.grid_columnconfigure(1, weight=1)

    container_frame = tk.Frame(training_window)
    container_frame.grid(row=0, column=0, sticky="nsew", padx=10, pady=10)
    container_frame.grid_rowconfigure(0, weight=1)
    container_frame.grid_columnconfigure(0, weight=1)
    container_frame.grid_columnconfigure(1, weight=1)

    # Left frame for Segmentation Network
    left_frame = tk.Frame(container_frame)
    left_frame.grid(row=0, column=0, sticky="nsew", padx=10)
    left_frame.grid_rowconfigure(99, weight=1)
    left_frame.grid_columnconfigure(0, weight=1)
    left_frame.grid_columnconfigure(1, weight=1)

    row = 0
    tk.Label(left_frame, text="Segmentation Network", font=("Arial", 16)).grid(row=row, column=0, columnspan=2, pady=10, sticky="nsew")
    row += 1

    def update_batch_combobox_segment():
        app_state.update_batch_select_combobox_segment(select_path=selection_var_segmentation.get())

    use_selected_files_var = tk.BooleanVar()
    tk.Checkbutton(left_frame, text="Use segmented files only", variable=use_selected_files_var).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1

    # Radio buttons for selection
    selection_var_segmentation = tk.StringVar(value="current_day")
    tk.Radiobutton(left_frame, text="Current Day", variable=selection_var_segmentation, value="current_day",
                   command=update_batch_combobox_segment).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(left_frame, text="Current Experiment", variable=selection_var_segmentation, value="current_experiment",
                   command=update_batch_combobox_segment).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(left_frame, text="Current Bird", variable=selection_var_segmentation, value="current_bird",
                   command=update_batch_combobox_segment).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    
    # Combobox for batch file selection
    batch_file_var = tk.StringVar()
    training_batch_combobox_segmentation = ttk.Combobox(left_frame,textvariable=batch_file_var)
    training_batch_combobox_segmentation.set("Select Batch File")
    app_state.training_window.training_batch_combobox_segmentation = training_batch_combobox_segmentation
    app_state.update_batch_select_combobox_segment(select_path = selection_var_segmentation.get())
    training_batch_combobox_segmentation.grid(row=2, column=1, sticky="ew")

    tk.Label(left_frame, text="Training Dataset Name: ").grid(row=row, column=0, sticky="w")
    dataset_name_entry = tk.Entry(left_frame)
    dataset_name_entry.grid(row=row, column=1, sticky="ew")
    dataset_name_entry.insert(0, "edit_seg_dataset_name")
    row += 1

    tk.Label(left_frame, text="Chunk Size: ").grid(row=row, column=0, sticky="w")
    chunk_size_entry = tk.Entry(left_frame, textvariable=app_state.train_segmentation_params['chunk_size'])
    chunk_size_entry.grid(row=row, column=1, sticky="ew")
    row += 1

    tk.Label(left_frame, text="Hist Size: ").grid(row=row, column=0, sticky="w")
    hist_size_entry = tk.Entry(left_frame, textvariable=app_state.train_segmentation_params['hist_size'])
    hist_size_entry.grid(row=row, column=1, sticky="ew")
    row += 1

    tk.Checkbutton(left_frame, text="Overlap chunks", variable=app_state.train_segmentation_params['overlap_chunks']).grid(row=row, column=0, sticky="w")
    row += 1

    tk.Button(
        left_frame, text="Create Training Dataset",
        command=lambda: start_create_segmentation_training_dataset(
            app_state, dataset_name_entry.get(), use_selected_files_var.get(), selection_var_segmentation.get(), batch_file_var.get(), bird_combobox, experiment_combobox, day_combobox, root
        )
    ).grid(row=row, column=0, columnspan=2, sticky="ew")
    row += 1

    tk.Label(left_frame, text="").grid(row=row, column=0, columnspan=2)
    row += 1

    training_dataset_var = tk.StringVar()
    training_dataset_combobox = ttk.Combobox(left_frame, textvariable=training_dataset_var)
    training_dataset_combobox.set("Select Training Dataset")
    app_state.training_window.training_dataset_combobox_segmentation = training_dataset_combobox
    app_state.update_segmentation_datasets_combobox()
    training_dataset_combobox.grid(row=row, column=0, columnspan=2, sticky="ew")
    row += 1

    checkbox_frame_segmentation = tk.Frame(left_frame)
    checkbox_frame_segmentation.grid(row=row, column=0, columnspan=2, sticky="w")
    tk.Checkbutton(checkbox_frame_segmentation, text="Downsampling", variable=app_state.train_segmentation_params['downsampling']).grid(row=0, column=0, sticky="w")
    #tk.Checkbutton(checkbox_frame_segmentation, text="QAT", variable=app_state.train_segmentation_params['qat']).grid(row=0, column=2, sticky="w") # optional for further usage
    row += 1

    segmentation_params = [
        ("Epochs:", 'epochs'),
        ("Batch Size:", 'batch_size'),
        ("Learning Rate:", 'learning_rate'),
        ("Early Stopping Patience:", 'early_stopping_patience'),
    ]

    for label_text, param_key in segmentation_params:
        tk.Label(left_frame, text=label_text).grid(row=row, column=0, sticky="w")
        tk.Entry(left_frame, textvariable=app_state.train_segmentation_params[param_key]).grid(row=row, column=1, sticky="ew")
        row += 1

    tk.Button(
        left_frame, text="Start Training",
        command=lambda: start_segmentation_training(root, app_state, training_dataset_var.get())
    ).grid(row=row, column=0, columnspan=2, sticky="ew")

    # Right frame for Classification Network
    right_frame = tk.Frame(container_frame)
    right_frame.grid(row=0, column=1, sticky="nsew", padx=10)
    right_frame.grid_rowconfigure(99, weight=1)
    right_frame.grid_columnconfigure(0, weight=1)
    right_frame.grid_columnconfigure(1, weight=1)

    row = 0
    tk.Label(right_frame, text="Classification Network", font=("Arial", 16)).grid(row=row, column=0, columnspan=2, pady=10, sticky="nsew")
    row += 1

    def update_batch_combobox_class():
        app_state.update_batch_select_combobox_class(select_path=selection_var_classification.get())

    use_selected_files_var_classification = tk.BooleanVar()
    tk.Checkbutton(right_frame, text="Use classified files only", variable=use_selected_files_var_classification).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1

    # Radio buttons for selection
    selection_var_classification = tk.StringVar(value="current_day")
    tk.Radiobutton(right_frame, text="Current Day", variable=selection_var_classification, value="current_day", 
                   command=update_batch_combobox_class).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(right_frame, text="Current Experiment", variable=selection_var_classification, value="current_experiment",
                   command=update_batch_combobox_class).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1
    tk.Radiobutton(right_frame, text="Current Bird", variable=selection_var_classification, value="current_bird",
                   command=update_batch_combobox_class).grid(row=row, column=0, columnspan=2, sticky="w")
    row += 1

    # Combobox for batch file selection
    batch_file_var = tk.StringVar()
    training_batch_combobox_classification = ttk.Combobox(right_frame,textvariable=batch_file_var)
    training_batch_combobox_classification.set("Select Batch File")
    app_state.training_window.training_batch_combobox_classification = training_batch_combobox_classification
    app_state.update_batch_select_combobox_class(select_path = selection_var_classification.get())
    training_batch_combobox_classification.grid(row=2, column=1, sticky="ew")

    tk.Label(right_frame, text="Training Dataset Name:").grid(row=row, column=0, sticky="w")
    dataset_name_entry_classification = tk.Entry(right_frame)
    dataset_name_entry_classification.grid(row=row, column=1, sticky="ew")
    dataset_name_entry_classification.insert(0, "edit_class_dataset_name")
    row += 1

    spec_params = [
        ("N Input Chunks / Size:", 'input_length'),
        ("Nperseg:", 'nperseg'),
        ("Noverlap:", 'noverlap'),
        ("NFFT:", 'nfft'),
        ("Frequency Cutoffs:", 'freq_cutoffs'),
    ]

    for label_text, param_key in spec_params:
        tk.Label(right_frame, text=label_text).grid(row=row, column=0, sticky="w")
        tk.Entry(right_frame, textvariable=app_state.spec_params[param_key]).grid(row=row, column=1, sticky="ew")
        row += 1

    tk.Button(
        right_frame, text="Create Training Dataset",
        command=lambda: start_create_classification_training_dataset(
            app_state, dataset_name_entry_classification.get(), use_selected_files_var_classification.get(), selection_var_classification.get(), batch_file_var.get(), bird_combobox, experiment_combobox, day_combobox, root
        )
    ).grid(row=row, column=0, columnspan=2, sticky="ew")
    row += 1

    tk.Label(right_frame, text="").grid(row=row, column=0, columnspan=2)
    row += 1

    training_dataset_var_classification = tk.StringVar()
    training_dataset_combobox_classification = ttk.Combobox(right_frame, textvariable=training_dataset_var_classification)
    training_dataset_combobox_classification.set("Select Training Dataset")
    app_state.training_window.training_dataset_combobox_classification = training_dataset_combobox_classification
    app_state.update_classification_datasets_combobox()
    training_dataset_combobox_classification.grid(row=row, column=0, columnspan=2, sticky="ew")
    row += 1

    checkbox_frame_classification = tk.Frame(right_frame)
    checkbox_frame_classification.grid(row=row, column=0, columnspan=2, sticky="w")
    tk.Checkbutton(checkbox_frame_classification, text="Downsampling", variable=app_state.train_classification_params['downsampling']).grid(row=0, column=0, sticky="w")
    #tk.Checkbutton(checkbox_frame_classification, text="QAT", variable=app_state.train_classification_params['qat']).grid(row=0, column=2, sticky="w") # make invisible for now
    row += 1

    classification_params = [
        ("Epochs:", 'epochs'),
        ("Batch Size:", 'batch_size'),
        ("Learning Rate:", 'learning_rate'),
        ("Early Stopping Patience:", 'early_stopping_patience'),
    ]

    for label_text, param_key in classification_params:
        tk.Label(right_frame, text=label_text).grid(row=row, column=0, sticky="w")
        tk.Entry(right_frame, textvariable=app_state.train_classification_params[param_key]).grid(row=row, column=1, sticky="ew")
        row += 1

    tk.Button(
        right_frame, text="Start Training",
        command=lambda: start_classification_training(
            root,
            app_state,
            training_dataset_var_classification.get(),
            bird_combobox.get()
        )
    ).grid(row=row, column=0, columnspan=2, sticky="ew")

    # Adjust window size to content
    training_window.update_idletasks()
    training_window.minsize(training_window.winfo_reqwidth(), training_window.winfo_reqheight())


def open_cluster_window(root, app_state, bird_combobox, experiment_combobox, day_combobox):
    """Open the cluster window for obtaining labels via clustering."""
    from moove.utils import start_create_cluster_dataset_thread, start_clustering_thread, start_dash_app_thread, replace_labels_from_df, stop_dash_app_thread, remove_pkl_suffix

    cluster_window = tk.Toplevel(root)
    cluster_window.title("Cluster")

    cluster_window.geometry("400x565")
    cluster_window.grid_rowconfigure(0, weight=1)
    cluster_window.grid_columnconfigure(0, weight=1)
    # ensure the window stays in front
    cluster_window.transient(root)

    app_state.cluster_window = cluster_window

    container_frame = tk.Frame(cluster_window)
    container_frame.grid(row=0, column=0, padx=10, pady=10, sticky="nsew")
    container_frame.grid_columnconfigure(0, weight=1)

    row_num = 0
    tk.Label(container_frame, text="Cluster Operations", font=("Arial", 16)).grid(row=row_num, column=0, columnspan=2, pady=10, sticky="ew")
    row_num += 1

    def update_batch_combobox_cluster():
        app_state.update_batch_select_combobox_cluster(select_path=selection_var.get())

    # CheckBox for "Use selected files only"
    use_selected_files_var = tk.BooleanVar()
    tk.Checkbutton(container_frame, text="Use segmented files only", variable=use_selected_files_var).grid(row=row_num, column=0, sticky="w")
    row_num += 1

    # Radio buttons for selection
    selection_var = tk.StringVar(value="current_day")
    tk.Radiobutton(container_frame, text="Current Day", variable=selection_var, value="current_day",
                   command=update_batch_combobox_cluster).grid(row=row_num, column=0, sticky="w")
    row_num += 1
    tk.Radiobutton(container_frame, text="Current Experiment", variable=selection_var, value="current_experiment",
                   command=update_batch_combobox_cluster).grid(row=row_num, column=0, sticky="w")
    row_num += 1
    tk.Radiobutton(container_frame, text="Current Bird", variable=selection_var, value="current_bird",
                   command=update_batch_combobox_cluster).grid(row=row_num, column=0, sticky="w")
    row_num += 1

    # Combobox for batch file selection
    batch_file_var = tk.StringVar()
    cluster_batch_combobox = ttk.Combobox(container_frame,textvariable=batch_file_var)
    cluster_batch_combobox.set("Select Batch File")
    app_state.cluster_window.cluster_batch_combobox = cluster_batch_combobox
    app_state.update_batch_select_combobox_cluster(select_path = selection_var.get())
    cluster_batch_combobox.grid(row=2, column=1, sticky="ew")

    # Entry for cluster dataset name
    tk.Label(container_frame, text="Cluster Dataset Name:").grid(row=row_num, column=0, sticky="w")
    dataset_name_entry = tk.Entry(container_frame)
    dataset_name_entry.grid(row=row_num, column=1, sticky="ew")
    dataset_name_entry.insert(0, "edit_cluster_dataset_name")
    row_num += 1

    spec_params = [
        ("Nperseg:", 'nperseg'),
        ("Noverlap:", 'noverlap'),
        ("NFFT:", 'nfft'),
        ("Frequency Cutoffs:", 'freq_cutoffs'),
    ]

    for label_text, param_key in spec_params:
        tk.Label(container_frame, text=label_text).grid(row=row_num, column=0, sticky="w")
        tk.Entry(container_frame, textvariable=app_state.spec_params[param_key]).grid(row=row_num, column=1, sticky="ew")
        row_num += 1

    # Button to create cluster dataset
    tk.Button(
        container_frame, text="Create Cluster Dataset",
        command=lambda: start_create_cluster_dataset_thread(app_state, dataset_name_entry.get(), use_selected_files_var.get(), selection_var.get(), batch_file_var.get(), bird_combobox, experiment_combobox, day_combobox, root)
    ).grid(row=row_num, column=0, columnspan=2, pady=(10, 0), sticky="ew")
    row_num += 1

    tk.Label(container_frame, text="").grid(row=row_num, column=0, columnspan=2)
    row_num += 1

    # ComboBox for selecting the cluster dataset
    cluster_dataset_var = tk.StringVar()
    cluster_dataset_combobox = ttk.Combobox(container_frame, textvariable=cluster_dataset_var)
    cluster_dataset_combobox.grid(row=row_num, column=0, columnspan=2, sticky="ew")
    cluster_dataset_combobox.set("Select Cluster Dataset")
    app_state.cluster_window.cluster_dataset_combobox = cluster_dataset_combobox
    app_state.update_cluster_datasets_combobox()
    row_num += 1

    umap_k_means_params = [
        ("N_neighbors:", 'n_neighbors'),
        ("Min_dist:", 'min_dist'),
        ("N Syllables:", 'n_clusters'),
    ]

    for label_text, param_key in umap_k_means_params:
        tk.Label(container_frame, text=label_text).grid(row=row_num, column=0, sticky="w")
        tk.Entry(container_frame, textvariable=app_state.umap_k_means_params[param_key]).grid(row=row_num, column=1, sticky="ew")
        row_num += 1

    # Button to start clustering
    tk.Button(
        container_frame, text="Cluster Syllables",
        command=lambda: start_clustering_thread(root, app_state, remove_pkl_suffix(cluster_dataset_var.get()))
    ).grid(row=row_num, column=0, columnspan=2, pady=(10, 0), sticky="ew")
    row_num += 1

    # Button to open Dash GUI
    tk.Button(
        container_frame, text="Open Dash GUI",
        command=lambda: start_dash_app_thread(app_state, remove_pkl_suffix(cluster_dataset_var.get()))
    ).grid(row=row_num, column=0, pady=(10, 0), sticky="ew")

    # Button to close Dash GUI
    tk.Button(
        container_frame, text="Close Dash GUI",
        command=lambda: stop_dash_app_thread(app_state)
    ).grid(row=row_num, column=1, pady=(10, 0), sticky="ew")

    row_num += 1

    # Button to replace labels
    tk.Button(
        container_frame, text="Replace Labels",
        command=lambda: replace_labels_from_df(app_state, remove_pkl_suffix(cluster_dataset_var.get()), root)
    ).grid(row=row_num, column=0, columnspan=2, pady=(10, 0), sticky="ew")

    # Adjust window size to content
    cluster_window.update_idletasks()
    cluster_window.minsize(cluster_window.winfo_reqwidth(), cluster_window.winfo_reqheight())



# utils/__init__.py
from .file_utils import get_directories, read_batch, get_file_data_by_index, get_display_data, save_seg_class_recfile, get_files_for_day, get_files_for_experiment, get_files_for_bird, filter_segmented_files, filter_classified_files, save_features, remove_pkl_suffix, find_batch_files, create_batch_file
from .plot_utils import update_plots, update_ax2_ax3, update_ax2, plot_data
from .audio_utils import seconds_to_index, index_to_seconds, decibel, seconds_to_chunk_index
from .syllable_utils import add_new_segment, select_event, edit_syllable, handle_keypress
from .gui_utils import zoom, unzoom, swipe_left, swipe_right, update
from .movefuncs_utils import handle_playback, handle_delete, handle_crop, extract_raw_audio, save_recfile, load_recfile, save_cbin, save_notmat
from .window_utils import open_resegment_window, open_cluster_window, open_training_window, open_relabel_window
from .segment_utils import start_segment_evfuncs, segment_ml, start_create_segmentation_training_dataset, start_segment_files_thread
from .label_utils import start_create_classification_training_dataset, create_classification_training_dataset, start_classify_files_thread 
from .training_utils import start_segmentation_training, start_classification_training 
from .clustering_utils import start_clustering_thread, replace_labels_from_df, start_create_cluster_dataset_thread
from .dash_utils import start_dash_app_thread, stop_dash_app_thread



import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import os
from moove.utils.audio_utils import (decibel)
from moove.utils.movefuncs_utils import (load_recfile)

plt.rcParams.update({
    'font.size': 14,
    'axes.labelsize': 14,
    'xtick.labelsize': 14,
    'ytick.labelsize': 14
})


def update_plots(display_dict, app_state, filepath):
    """Update plots with new display data."""
    ax1, ax2, ax3 = app_state.get_axes()

    for ax in (ax1, ax2, ax3): 
        ax.tick_params(axis='both', colors=app_state.text_color) 

    plt.tight_layout() 
    plt.subplots_adjust(left=0.129, right=0.999, top=0.92, bottom=0.1, hspace=0.1) 

    vmin, vmax = app_state.current_vmin, app_state.current_vmax

    # Filter frequencies
    freqs = display_dict["freqs"]
    valid_freqs = (freqs >= app_state.config['lower_spec_plot']) & (freqs <= app_state.config['upper_spec_plot'])
    filtered_freqs = freqs[valid_freqs]
    filtered_spectrogram_data = display_dict["spectrogram_data"][valid_freqs, :]
    # calculates the power into amplitudes
    amplitude_spec = np.sqrt(filtered_spectrogram_data)
    # amplitude -> dB
    db_spec = decibel(amplitude_spec)

    feedback_time = []
    # Load info from .rec file
    catch = load_recfile(os.path.splitext(filepath["file_path"])[0] + ".rec")["catch_song"]
    feedback_time = load_recfile(os.path.splitext(filepath["file_path"])[0] + ".rec")["feedback_info"]

    # Update ax1 (Spectrogram)
    ax1.clear()

    performance_mode = app_state.config['performance']
    
    if performance_mode == "fast":
        # faster, less detailed drawings
        extent = [display_dict["times"].min(), display_dict["times"].max(),
                filtered_freqs.min(), filtered_freqs.max()]

        app_state.spec = ax1.imshow(db_spec, aspect='auto', origin='lower',
                                    extent=extent, cmap='jet', vmin=vmin, vmax=vmax)
    else:
        # slower, more detailed drawing
        app_state.spec = ax1.pcolormesh(display_dict["times"], filtered_freqs, db_spec,
                                        shading='gouraud', cmap='jet', vmin=vmin, vmax=vmax)

    # get axis position in figure coordinates
    pos = ax2.get_position()
    fig = ax2.figure
    # remove all previous figure level text
    for txt in list(fig.texts):
        txt.remove()
    fig.patches = [p for p in fig.patches if not (isinstance(p, patches.Rectangle) and getattr(p, 'is_box', False))]

    # Mark catch trials in GUI
    if catch == 1:
        ax1.text(0.0, 1.0, "catch", color='crimson', fontsize=17, transform=ax1.transAxes)
        rect = patches.Rectangle((pos.x0, pos.y0), pos.width, pos.height, linewidth=1,
                                 edgecolor='crimson', facecolor='none', transform=fig.transFigure, zorder=100)
        rect.is_box = True
        fig.patches.append(rect)

    # Mark feedback timings
    feedback_timings = []
    if feedback_time:
        for i in range(len(feedback_time)):
            feedback_timings.append(feedback_time[i][0])
        for t in feedback_timings:
            ax1.text(t, -0.01, '^', ha='center', va='top', color='red', fontsize=24,
                     transform=ax1.get_xaxis_transform(), zorder=1000000)

    ax1.set_title(display_dict["file_name"], color=app_state.text_color)
    ax1.set_ylabel('Frequency (Hz)', color=app_state.text_color)

    # Update ax2 (Label visualization)
    ax2.clear()
    if "labels" in display_dict:
        min_len = min(len(display_dict["onsets"]), len(display_dict["offsets"]), len(display_dict["labels"]))
        for i, label in zip(range(min_len), display_dict["labels"]):
            label_position = (display_dict["onsets"][i] + display_dict["offsets"][i]) / (2 * 1000)
            ax2.text(label_position, 0.5, label, ha='center', va='center', clip_on=True)

        app_state.logger.debug("***")
        app_state.logger.debug("display_dict onsets: %s", display_dict["onsets"])
        app_state.logger.debug("display_dict offsets: %s", display_dict["offsets"])
        app_state.logger.debug("display_dict labels: %s", display_dict["labels"])

    ax2.set_yticks([])
    ax2.set_xlim(display_dict["times"][0], display_dict["times"][-1])

    # Update ax3 (Amplitude plot)
    ax3.clear()
    ax3.plot(np.arange(len(display_dict["song_data"])) / display_dict["sampling_rate"], display_dict["amplitude"],
             color='#2653c5')
    ax3.set_ylabel('Amplitude (dB)', color=app_state.text_color)
    ax3.set_xlabel('Time (s)', color=app_state.text_color)
    ax3.set_ylim(display_dict["amplitude"].min(), display_dict["amplitude"].max())

    # Mark onsets and offsets
    if "onsets" in display_dict and "offsets" in display_dict:
        bar_height = float(app_state.evfuncs_params['threshold'].get())
        for onset, offset in zip(display_dict["onsets"], display_dict["offsets"]):
            onset_sec = onset / 1000
            offset_sec = offset / 1000
            ax3.hlines(bar_height, onset_sec, offset_sec, color='black', linewidth=1.5)
            ax3.plot(onset_sec, bar_height, marker='+', color='black', markersize=10, markeredgewidth=1.5)
            ax3.plot(offset_sec, bar_height, marker='+', color='black', markersize=10, markeredgewidth=1.5)

    app_state.set_axes(ax1, ax2, ax3)

    fig.patch.set_facecolor(app_state.bg_color) 

    fig.align_ylabels() 
    app_state.draw_canvas()


def update_ax2_ax3(ax2, ax3, display_dict, app_state):
    """Update the second and third axes with new display data."""
    app_state.logger.debug("Updating ax2 and ax3")

    x_lim = ax2.get_xlim()
    app_state.logger.debug("Current x-axis limits: %s", x_lim)

    ax2.clear()
    # Update ax2 with labels
    if "labels" in display_dict:
        for i, label in enumerate(display_dict["labels"]):
            label_position = (display_dict["onsets"][i] + display_dict["offsets"][i]) / (2 * 1000)
            ax2.text(label_position, 0.5, label, ha='center', va='center', clip_on=True)
    ax2.set_yticks([])
    ax2.set_xlim(x_lim)

    ax3.clear()
    ax3.plot(np.arange(len(display_dict["song_data"])) / display_dict["sampling_rate"], display_dict["amplitude"],
             color='#2653c5')
    ax3.set_ylabel('Amplitude (dB)', color=app_state.text_color)
    ax3.set_xlabel('Time (s)', color=app_state.text_color)
    ax3.set_ylim(display_dict["amplitude"].min(), display_dict["amplitude"].max())
    ax3.set_xlim(x_lim)

    # Mark onsets and offsets
    if "onsets" in display_dict and "offsets" in display_dict:
        bar_height = float(app_state.evfuncs_params['threshold'].get())
        for onset, offset in zip(display_dict["onsets"], display_dict["offsets"]):
            onset_sec = onset / 1000
            offset_sec = offset / 1000
            ax3.hlines(bar_height, onset_sec, offset_sec, color='black', linewidth=1.5)
            ax3.plot(onset_sec, bar_height, marker='+', color='black', markersize=10, markeredgewidth=1.5)
            ax3.plot(offset_sec, bar_height, marker='+', color='black', markersize=10, markeredgewidth=1.5)

    app_state.draw_canvas()


def update_ax2(ax2, display_dict, app_state):
    """Update the second axis with new display data."""
    x_lim = ax2.get_xlim()
    y_lim = ax2.get_ylim()
    ax2.clear()
    ax2.set_yticks([])
    ax2.set_xlim(x_lim)

    # Cover up old labels temporarily
    ax2.add_patch(
        plt.Rectangle(
            (x_lim[0], y_lim[0]),
            x_lim[1] - x_lim[0],
            y_lim[1] - y_lim[0],
            color='white', zorder=0
        )
    )

    # write new labels to axis
    if "labels" in display_dict:
        for i in range(len(display_dict["onsets"])):
            label_position = (display_dict["onsets"][i] + display_dict["offsets"][i]) / 2000
            ax2.text(label_position, 0.5, display_dict["labels"][i], ha='center', va='center', clip_on=True, zorder=1)

    for artist in ax2.texts + ax2.patches:
        ax2.draw_artist(artist)
    app_state.canvas.blit(ax2.bbox)


def plot_data(app_state):
    """Plot new data and update the application state."""
    from moove.utils.file_utils import get_file_data_by_index, get_display_data

    try:
        file_path = get_file_data_by_index(app_state.data_dir, app_state.song_files, app_state.current_file_index, app_state)
        app_state.display_dict = get_display_data(file_path, app_state.config)
        
        update_plots(app_state.display_dict, app_state, file_path)
        ax1, ax2, ax3 = app_state.get_axes()
        ax1.set_navigate(False)

        # Set original axis ranges for zooming and unzooming
        original_x_range = (ax1.get_xlim()[0], ax1.get_xlim()[1])
        app_state.set_original_x_range(original_x_range)
        original_y_range_ax1 = (ax1.get_ylim()[0], ax1.get_ylim()[1])
        original_y_range_ax2 = (ax2.get_ylim()[0], ax2.get_ylim()[1])
        original_y_range_ax3 = (ax3.get_ylim()[0], ax3.get_ylim()[1])
        app_state.set_original_y_range_ax1(original_y_range_ax1, original_y_range_ax2, original_y_range_ax3)

        # Load and set segmented and classified checkboxes
        hand_segmented = load_recfile(os.path.splitext(file_path["file_path"])[0] + ".rec")["hand_segmented"]
        hand_classified = load_recfile(os.path.splitext(file_path["file_path"])[0] + ".rec")["hand_classified"]
        app_state.segmented_var.set(str(hand_segmented))
        app_state.classified_var.set(str(hand_classified))
        app_state.edit_type = "None"

        app_state.logger.debug("Recfile loaded and checkboxes updated for file: %s", file_path["file_name"])

        # Store the last valid file path for fallback
        if hasattr(app_state, 'last_valid_file_path'):
            app_state.last_valid_file_path = file_path["file_path"]
        else:
            app_state.last_valid_file_path = file_path["file_path"]
        
        app_state.reset_edit_type_gui()

        app_state.draw_canvas()
        
    except:
        # If there's an error with the current file, try to fall back to the last valid file
        app_state.logger.debug("Could not plot data. Trying to fall back to last valid file.")
        
        if hasattr(app_state, 'last_valid_file_path') and app_state.last_valid_file_path:
            try:
                # Try to plot the last valid file instead
                fallback_file_data = {"file_name": os.path.basename(app_state.last_valid_file_path), 
                                    "file_path": app_state.last_valid_file_path}
                app_state.display_dict = get_display_data(fallback_file_data, app_state.config)
                
                update_plots(app_state.display_dict, app_state, fallback_file_data)
                ax1, ax2, ax3 = app_state.get_axes()
                ax1.set_navigate(False)

                # Set original axis ranges for zooming and unzooming
                original_x_range = (ax1.get_xlim()[0], ax1.get_xlim()[1])
                app_state.set_original_x_range(original_x_range)
                original_y_range_ax1 = (ax1.get_ylim()[0], ax1.get_ylim()[1])
                original_y_range_ax2 = (ax2.get_ylim()[0], ax2.get_ylim()[1])
                original_y_range_ax3 = (ax3.get_ylim()[0], ax3.get_ylim()[1])
                app_state.set_original_y_range_ax1(original_y_range_ax1, original_y_range_ax2, original_y_range_ax3)

                # Load and set segmented and classified for the checkboxes
                hand_segmented = load_recfile(os.path.splitext(fallback_file_data["file_path"])[0] + ".rec")["hand_segmented"]
                hand_classified = load_recfile(os.path.splitext(fallback_file_data["file_path"])[0] + ".rec")["hand_classified"]
                app_state.segmented_var.set(str(hand_segmented))
                app_state.classified_var.set(str(hand_classified))
                app_state.edit_type = "None"

                app_state.reset_edit_type_gui()

                app_state.logger.debug("Successfully fell back to last valid file: %s", fallback_file_data["file_name"])
                app_state.draw_canvas()
                
            except Exception as fallback_error:
                app_state.logger.debug("Failed to fall back to last valid file: %s", str(fallback_error))
                # If even the fallback fails, just log the error and continue
        else:
            app_state.logger.debug("No fallback file available. User will need to manually navigate to a valid file.")



# utils/label_utils.py
import os
import pandas as pd
import pickle
import threading
import tkinter as tk
import tkinter.font as tkFont
import torch
import torch.nn.functional as F
import evfuncs
import re
from scipy.signal import spectrogram
from tkinter import ttk, messagebox
import numpy as np


def load_classification_checkmarks(all_files):
    """Check whether files have been manually checked as being classified"""
    unclass_files = []
    for file in all_files:
        recfile_path = os.path.splitext(file)[0] + ".rec"
        with open(recfile_path, "r") as f:
            content = f.read()

        hand_segmented_pattern = r"Hand Classified = (\d+)"
        hand_segmented_match = re.search(hand_segmented_pattern, content)

        if hand_segmented_match.group(1) == '0':
            unclass_files.append(file)

    return unclass_files


def start_create_classification_training_dataset(app_state, dataset_name, use_selected_files, selection, batch_file, bird_combobox, experiment_combobox, day_combobox, root):
    """Initialize the creation of a classification training dataset in a new thread."""
    from moove.utils import get_files_for_day, get_files_for_experiment, get_files_for_bird, filter_classified_files
    
    font_style = tkFont.Font(family="Arial", size=14)
    running_label = tk.Label(app_state.training_window, text="Looking for files...", fg="green", font=font_style)
    running_label.grid(row=22, column=0, columnspan=2, pady=(10, 0), sticky=tk.W) 
    root.update_idletasks() 

    # choose files depending on the user selection
    bird, experiment, day = bird_combobox.get(), experiment_combobox.get(), day_combobox.get()
    if selection == "current_day":
        files = get_files_for_day(app_state, bird, experiment, day, batch_file)
    elif selection == "current_experiment":
        files = get_files_for_experiment(app_state, bird, experiment, batch_file)
    elif selection == "current_bird":
        files = get_files_for_bird(app_state, bird, batch_file)

    if use_selected_files:
        files = filter_classified_files(files)
        
    running_label.destroy()
    root.update_idletasks() 

    dataset_name = str(dataset_name)
    # check if dataset name is valid
    if len(dataset_name) < 1:
        messagebox.showinfo("Error", "Dataset name not valid! A dataset name needs to contain at least one character.")
    else:
        # create dataset with valid name
        max_value = len(files)
        progressbar = ttk.Progressbar(app_state.training_window, orient=tk.HORIZONTAL, length=200, mode='determinate', maximum=max_value)
        progressbar.grid(row=22, column=0, columnspan=2, pady=(10, 0), sticky="ew")

        # Create wrapper function for thread management
        def thread_wrapper():
            current_thread = threading.current_thread()
            try:
                create_classification_training_dataset(app_state, progressbar, dataset_name, files, root)
            finally:
                app_state.remove_thread(current_thread)
        
        thread = threading.Thread(target=thread_wrapper, name="CreateClassDatasetThread")
        app_state.add_thread(thread)
        thread.start()


def create_classification_training_dataset(app_state, progressbar, dataset_name, files, root):
    """Create a classification training dataset based on selected files and parameters."""
    from moove.utils import get_display_data, seconds_to_index

    if len(files) == 0:
        messagebox.showinfo("Error", "Not enough files given! You need at least 1 file to create a dataset.")
        return

    input_length_str = app_state.spec_params['input_length'].get()
    input_length, chunk_size = map(int, input_length_str.split(','))
    nperseg, noverlap, nfft = int(app_state.spec_params['nperseg'].get()), int(app_state.spec_params['noverlap'].get()), int(app_state.spec_params['nfft'].get())
    freq_cutoffs = tuple(map(int, app_state.spec_params['freq_cutoffs'].get().split(',')))
    input_array_size = input_length * chunk_size
    
    going_prod_df = pd.DataFrame(columns=['file', 'onset_no', 'taf_unflattend_spectrogram', 'label'])
    entry_no = 0

    progressbar.grid_remove()

    # Add running label to the GUI
    font_style = tkFont.Font(family="Arial", size=14)
    running_label = tk.Label(app_state.training_window, text="Looking for syllables...", fg="green", font=font_style)
    running_label.grid(row=22, column=0, columnspan=2, pady=(10, 0), sticky=tk.W) 
    root.update_idletasks() 

    def get_onsets(file_path):
        notmat_file = file_path + ".not.mat"
        if os.path.exists(notmat_file):
            notmat_dict = evfuncs.load_notmat(notmat_file)
            return notmat_dict.get("onsets", [])
        else:
            return []
    
    # Count number of syllable onsets    
    num_onsets = 0
    for i, file_i in enumerate(files):
        working_dir = os.getcwd()
        file_path = os.path.join(working_dir, file_i)
        onsets = get_onsets(file_path)
        if len(onsets) > 0:
            num_onsets += len(onsets)

    if num_onsets == 0:
        running_label.destroy()
        root.update_idletasks() 
        messagebox.showinfo("Error", "No syllable onsets found in the given files.")
        return

    running_label.destroy()
    root.update_idletasks() 
    progressbar.grid()

    for i, file_i in enumerate(files):
        app_state.training_window.update_idletasks()
        working_dir = os.getcwd()
        file_path = os.path.join(working_dir, file_i)
        file_data = get_display_data({"file_name": os.path.basename(file_path), "file_path": file_path}, app_state.config)
        sampling_rate = int(file_data["sampling_rate"])
        rawsong, onsets, labels = file_data["song_data"], file_data["onsets"], file_data["labels"]

        if len(onsets) > 0:
            progressbar['value'] = i
            for syllable_no, onset in enumerate(onsets):
                entry_no += 1
                onset_index = int(seconds_to_index(onset, sampling_rate))
                cutted_raw_song = rawsong[onset_index:onset_index + input_array_size]

                # Ensure consistent shape by setting nperseg and noverlap
                f, t, Sxx_taf = spectrogram(cutted_raw_song, fs=sampling_rate, nperseg=nperseg, noverlap=noverlap, nfft=nfft)
                if Sxx_taf.ndim == 2:
                    Sxx_taf = Sxx_taf[(f >= freq_cutoffs[0]) & (f <= freq_cutoffs[1]), :]
                else:
                    app_state.logger.warning(f"Warning: Sxx_taf is {Sxx_taf.ndim}-dimensional for file {file_i}, skipping this entry.")
                    continue

                going_prod_df.loc[entry_no] = [file_i, syllable_no, Sxx_taf, labels[syllable_no]]

    metadata = {
        'input_length': input_length_str,
        'nperseg': nperseg,
        'noverlap': noverlap,
        'nfft': nfft,
        'lowcut': freq_cutoffs[0],
        'highcut': freq_cutoffs[1],
    }

    save_path = os.path.join(app_state.config['global_dir'], 'training_data', f'{dataset_name}_class.pkl')
    with open(save_path, 'wb') as f:
        pickle.dump({'dataframe': going_prod_df, 'metadata': metadata}, f)

    app_state.update_classification_datasets_combobox()
    progressbar['value'] = len(files)
    progressbar.grid_forget()
    
    # Schedule Tkinter operations in the main thread
    def show_message():
        messagebox.showinfo("Info", "Classification training dataset has been created successfully!")
    
    root.after(0, show_message)

    first_index = going_prod_df.index[0]
    shape_first_entry = pd.DataFrame(going_prod_df.loc[first_index, 'taf_unflattend_spectrogram']).shape
    app_state.logger.debug(f"The shape of the first entry in 'taf_unflattend_spectrogram' is {shape_first_entry}")


def normalize_spectrogram(spectrogram):
    """Normalize the spectrogram to zero mean and unit variance."""
    mean, std = spectrogram.mean(), spectrogram.std()
    return (spectrogram - mean) / std if std != 0 else spectrogram


def start_classify_files_thread(app_state, model_name, selection, checkbox_ow, batch_file, bird_combobox, experiment_combobox, day_combobox):
    """Start the classification process for selected files in a new thread."""
    from moove.utils import get_files_for_day, get_files_for_experiment, get_files_for_bird, get_file_data_by_index

    bird, experiment, day = bird_combobox.get(), experiment_combobox.get(), day_combobox.get()
    # Choose files depending on the user selection
    if selection == "current_day":
        files = get_files_for_day(app_state, bird, experiment, day, batch_file)
    elif selection == "current_experiment":
        files = get_files_for_experiment(app_state, bird, experiment, batch_file)
    elif selection == "current_bird":
        files = get_files_for_bird(app_state, bird, batch_file)
    elif selection == "current_file":
        files = [get_file_data_by_index(app_state.data_dir, app_state.song_files, app_state.current_file_index, app_state)["file_path"]]

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    try:
        checkpoint = torch.load(os.path.join(app_state.config['global_dir'], 'trained_models', f'{model_name}.pth'), map_location=device)
    except: 
        messagebox.showinfo("Error", "Selected classification model doesn't exist or is not valid! Perhaps you forgot to pick a model?")
    model, metadata = checkpoint['model'], checkpoint['metadata']

    model.to(device).eval()

    if checkbox_ow:
        files = files
    else:
        # use non-manually classified files only
        files = load_classification_checkmarks(files)

    max_value = len(files)
    progressbar = ttk.Progressbar(app_state.relabel_window, orient=tk.HORIZONTAL, length=200, mode='determinate', maximum=max_value)
    progressbar.grid(row=22, column=0, columnspan=2, pady=(10, 0), sticky="ew")
    
    # Create wrapper function for thread management
    def thread_wrapper():
        current_thread = threading.current_thread()
        try:
            ml_classify_file(app_state, progressbar, max_value, files, model, metadata, device)
        finally:
            app_state.remove_thread(current_thread)
    
    thread = threading.Thread(target=thread_wrapper, name="ClassifyFilesThread")
    app_state.add_thread(thread)
    thread.start()


def ml_classify_file(app_state, progressbar, max_value, all_files, model, metadata, device):
    """Perform classification on each file and update labels."""
    from moove.utils import get_display_data, plot_data, save_notmat, seconds_to_index

    # Store the complete original file context to restore it later
    original_data_dir = app_state.data_dir
    original_song_files = app_state.song_files.copy() if app_state.song_files else []
    original_current_file_index = app_state.current_file_index

    input_length, chunk_size = map(int, metadata['input_length'].split(','))
    nperseg, noverlap, nfft = int(metadata['nperseg']), int(metadata['noverlap']), int(metadata['nfft'])
    lowcut, highcut, int_to_label = int(metadata['lowcut']), int(metadata['highcut']), metadata['int_to_label']
    input_array_size = input_length * chunk_size

    for i, file_i in enumerate(all_files):
        try:
            progressbar['value'] = i
            app_state.relabel_window.update_idletasks()
            file_data = get_display_data({"file_name": os.path.basename(file_i), "file_path": file_i}, app_state.config)
            sampling_rate, rawsong, onsets = int(file_data["sampling_rate"]), file_data["song_data"], file_data["onsets"]
            app_state.data_dir = os.path.dirname(file_i)

            labels = []

            # Classify syllables in selected files
            for onset in onsets:
                onset_index = int(seconds_to_index(onset, sampling_rate))
                cutted_raw_song = rawsong[onset_index:onset_index + input_array_size]

                f, _, Sxx_taf = spectrogram(cutted_raw_song, fs=sampling_rate, nperseg=nperseg, noverlap=noverlap, nfft=nfft)
                Sxx_taf = Sxx_taf[(f >= lowcut) & (f <= highcut), :]
                Sxx_normalized = normalize_spectrogram(Sxx_taf)

                input_data = torch.tensor(Sxx_normalized).float().unsqueeze(0).unsqueeze(0).to(device)
                input_tensor = F.pad(input_data, (0, 1, 0, 1))

                with torch.no_grad():
                    output = model(input_tensor)
                    predicted_class = torch.argmax(output).item()
                labels.append(int_to_label[predicted_class])

            file_data["labels"] = ''.join(labels)
            save_notmat(os.path.join(app_state.data_dir, f"{file_data['file_name']}.not.mat"), file_data)
            
        except Exception as e:
            app_state.info(f"File {file_i} could not be processed correctly: {e}. Check manually.")
            return

    # Restore the complete original file context
    app_state.data_dir = original_data_dir
    app_state.song_files = original_song_files
    app_state.current_file_index = original_current_file_index

    progressbar['value'] = len(all_files)
    
    # Final UI update
    app_state.reset_edit_type()
    plot_data(app_state)
    progressbar.grid_forget()
    
    # Schedule Tkinter operations in the main thread
    def show_message():
        messagebox.showinfo("Info", f"Relabeling of files completed successfully!")
    
    app_state.relabel_window.after(0, show_message)



# utils/file_utils.py
import evfuncs
import os
import re
import pickle
from scipy.io import wavfile as wav
from scipy.signal import spectrogram
from tkinter import messagebox
from moove.utils.audio_utils import decibel


def get_directories(path):
    """Return a list of directories in the specified path."""
    return [d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))]


def create_batch_file(data_dir):
    """Create a default batch.txt file."""
    # scan the directory for .wav and .cbin files
    valid_files = [f for f in os.listdir(data_dir) if f.endswith('.wav') or f.endswith('.cbin')]

    # path to default batch file
    batch_path = os.path.join(data_dir, 'batch.txt')

    # open batch file for writing
    with open(batch_path, 'w') as batch_file:
        for audio_file in valid_files:
            # write each .wav and .cbin file name to batch file
            batch_file.write(f"{audio_file}\n")

    print(f"Default batch.txt file created with {len(valid_files)} entries.")


def find_batch_files(day_path):
    """Find all files in the day directory that match the pattern '.*batch*.' with allowed extensions only"""
    import re
    
    # Define allowed extensions for batch files
    allowed_extensions = ['', '.txt', '.keep']  # No extension, .txt, .keep
    
    all_files = os.listdir(day_path)
    batch_files = []
    
    for f in all_files:
        # Check if filename contains 'batch'
        if re.search(r'batch', f, re.IGNORECASE):
            # Get file extension
            _, ext = os.path.splitext(f)
            
            # Only include files with allowed extensions
            if ext in allowed_extensions:
                batch_files.append(f)
    
    # create default batch files if non found/ no default'batch.txt' found
    if not batch_files:
        create_batch_file(day_path)
        batch_files = ['batch.txt']  # Add the newly created file to the list
    if 'batch.txt' not in batch_files:
        create_batch_file(day_path)     
        # ensure default 'batch' is first   
        batch_files.insert(0, 'batch.txt') 
    return batch_files


def read_batch(day_path, batch_file="batch.txt"):
    """Read and return the contents of the specified batch file as a list of lines."""
    file_path = os.path.join(day_path, batch_file)
    with open(file_path, "r") as file:
        content = file.readlines()
    return [line.strip() for line in content]


def remove_line(file_path, rm_line):
    """Remove line from a given batch file."""
    # read all lines from the file
    with open(file_path, 'r') as file:
        lines = file.readlines()

    # remove the line (strip newline for accurate match)
    lines = [line for line in lines if line.strip() != rm_line]

    # write the filtered lines back to the file
    with open(file_path, 'w') as file:
        file.writelines(lines)


def get_file_data_by_index(path, song_files, file_index, app_state):
    """Retrieve file data by index from the app state.
       Delete entry from song list and all batch files if song file is missing."""
    try:
        current_file = song_files[file_index]
    except IndexError:
        app_state.current_file_index = 0
        print(f"File not found, entry removed - defaulting to first file.")
        current_file = song_files[0]
       
    file_path = os.path.join(os.getcwd(), path, current_file)

    file_data_dict = {"file_name": current_file, "file_path": file_path}
    return file_data_dict


def get_display_data(file_data_dict, config):
    """Generate display data including spectrogram and amplitude for a given file."""
    dspec_config = {
        "spec_nperseg": 1024,
        "spec_noverlap": 896,
        "spec_nfft": 1024,
    }

    if config is not None:
        spectrogram_config = {**dspec_config, **config}
    else:
        spectrogram_config = dspec_config

    file_name = file_data_dict["file_name"]
    file_path = file_data_dict["file_path"].replace("wsl$", "wsl.localhost")
    notmat_dict = {}

    if file_name.endswith(".cbin"):
        song_data, sampling_rate = evfuncs.load_cbin(file_path)
        if os.path.exists(file_path + ".not.mat"):
            notmat_dict = evfuncs.load_notmat(file_path + ".not.mat")
    elif file_name.endswith(".wav"):
        sampling_rate, song_data = wav.read(file_path)
        if os.path.exists(file_path + ".not.mat"):
            notmat_dict = evfuncs.load_notmat(file_path + ".not.mat")
    else:
        raise ValueError("Unsupported file format")

    smoothed_song_data = evfuncs.smooth_data(song_data, sampling_rate, (500, 10000), 2)
    freqs, times, spectrogram_data = spectrogram(song_data, 
                                                 fs=sampling_rate,
                                                 nperseg=spectrogram_config["spec_nperseg"],
                                                 noverlap=spectrogram_config["spec_noverlap"],
                                                 nfft=spectrogram_config["spec_nfft"]
    )
    amplitude = decibel(smoothed_song_data)

    display_dict = {
        "file_name": file_name,
        "sampling_rate": sampling_rate,
        "song_data": song_data,
        "freqs": freqs,
        "times": times,
        "spectrogram_data": spectrogram_data,
        "amplitude": amplitude,
    }

    if notmat_dict:
        display_dict.update(notmat_dict)
    
    return display_dict


def save_seg_class_recfile(filepath, segmented, classified):
    """Save or update recfile for a specific file"""
    try:
        with open(filepath, 'r') as f:
            lines = f.readlines()

        new_lines = []
        for line in lines:
            if line.strip().startswith('Hand Segmented ='):
                new_lines.append(f'Hand Segmented = {segmented}\n')
            elif line.strip().startswith('Hand Classified ='):
                new_lines.append(f'Hand Classified = {classified}\n')
            else:
                new_lines.append(line)

        with open(filepath, 'w') as f:
            f.writelines(new_lines)
            
    except FileNotFoundError:
        # If the recfile doesn't exist, just continue without saving
        # This prevents crashes when the file is in a different directory
        pass


def get_files_for_day(app_state, bird, experiment, day, batch_file="batch.txt"):
    """Return files for a specific day (and batch) within the experiment and bird directory."""
    day_path = os.path.join(app_state.config['rec_data'], bird, experiment, day)
    if batch_file == "All Files":
        # If "All Files" is selected, take all wav/cbin files from this day
        files = [f for f in os.listdir(day_path)
                 if f.lower().endswith('.wav') or f.lower().endswith('.cbin')]
    else:
        # Otherwise, read the batch file and only take files listed in it
        files = read_batch(day_path, batch_file)
    return [os.path.join(day_path, f) for f in files]


def get_files_for_experiment(app_state, bird, experiment, batch_file="batch.txt"):
    """Return all files for a specific experiment (and batch) of the specified bird."""
    experiment_path = os.path.join(app_state.config['rec_data'], bird, experiment)
    all_files = []
    
    days = get_directories(experiment_path)
    for day in days:
        day_path = os.path.join(app_state.config['rec_data'], bird, experiment, day)
        
        if batch_file == "All Files":
            # If "All Files" is selected, take all wav/cbin files from this day
            files_in_day = [f for f in os.listdir(day_path)
                            if f.lower().endswith('.wav') or f.lower().endswith('.cbin')]
            for file in files_in_day:
                full_file_path = os.path.join(day_path, file)
                all_files.append(full_file_path)
        else:
            # Otherwise, read the batch file and only take files listed in it
            files_in_batch = read_batch(experiment_path, batch_file)
            files_in_day = [f for f in os.listdir(day_path)
                            if f.lower().endswith('.wav') or f.lower().endswith('.cbin')]
            
            # Check which files from this day are in the batch file
            for file in files_in_day:
                if file in files_in_batch:
                    # Add the full path to the file
                    full_file_path = os.path.join(day_path, file)
                    all_files.append(full_file_path)
    
    return all_files


def get_files_for_bird(app_state, bird, batch_file="batch.txt"):
    """Return all files for a specific bird (and batch) from bird folder."""
    bird_path = os.path.join(app_state.config['rec_data'], bird)
    all_files = []
    
    # Get all experiment folders in the bird folder
    experiments = get_directories(bird_path)
    for experiment in experiments:
        experiment_path = os.path.join(bird_path, experiment)
        
        # Get all day folders in each experiment folder
        days = get_directories(experiment_path)
        for day in days:
            day_path = os.path.join(experiment_path, day)
            
            if batch_file == "All Files":
                # If "All Files" is selected, take all wav/cbin files from this day
                files_in_day = [f for f in os.listdir(day_path)
                                if f.lower().endswith('.wav') or f.lower().endswith('.cbin')]
                for file in files_in_day:
                    full_file_path = os.path.join(day_path, file)
                    all_files.append(full_file_path)
            else:
                # Otherwise, read the batch file and only take files listed in it
                files_in_batch = read_batch(bird_path, batch_file)
                files_in_day = [f for f in os.listdir(day_path)
                                if f.lower().endswith('.wav') or f.lower().endswith('.cbin')]
                
                # Check which files from this day are in the batch file
                for file in files_in_day:
                    if file in files_in_batch:
                        # Add the full path to the file
                        full_file_path = os.path.join(day_path, file)
                        all_files.append(full_file_path)
    
    return all_files


def filter_segmented_files(files):
    """Filter and return files that are segmented according to the recfile."""
    segment_those_files = []
    for file in files:
        recfile_path = os.path.splitext(file)[0]+".rec"
        with open(recfile_path, "r") as f:
            content = f.read()

        hand_segmented_pattern = r"Hand Segmented = (\d+)"
        hand_segmented_match = re.search(hand_segmented_pattern, content)

        if hand_segmented_match.group(1) == '1':
            segment_those_files.append(file)
            
    print(f"Total segmented files found: {len(segment_those_files)}")
    
    return segment_those_files


def filter_classified_files(files):
    """Filter and return files that are classified according to the recdata."""
    classify_those_files = []
    for file in files:
        recfile_path = os.path.splitext(file)[0] + ".rec"
        with open(recfile_path, "r") as f:
            content = f.read()

        hand_segmented_pattern = r"Hand Classified = (\d+)"
        hand_segmented_match = re.search(hand_segmented_pattern, content)

        if hand_segmented_match.group(1) == '1':
            classify_those_files.append(file)

    print(f"Total classified files found: {len(classify_those_files)}")
    
    return classify_those_files


def save_features(app_state, dataset_name, arr, chunk_size=64, hist_size=3, num_files=None, num_syls = None):
    """Save feature arrays to a dataset with the specified chunk and history size metadata."""
    metadata = {
        'chunk_size': chunk_size,
        'hist_size': hist_size
    }
    
    save_path = os.path.join(app_state.config['global_dir'], 'training_data', f'{dataset_name}_seg.pkl')
    if num_files is None or num_files <= 1:
        # Save feature array and metadata in a single file
        with open(save_path, 'wb') as f:
            pickle.dump({'features': arr, 'metadata': metadata, 'syllables': num_syls}, f)
    else:
        # Split and save the array into multiple files
        split_size = arr.shape[0] // num_files
        for i in range(num_files):
            start_idx = i * split_size
            end_idx = (i + 1) * split_size if i < num_files - 1 else arr.shape[0]
            with open(f'training_data/{dataset_name}_{i}_seg.pkl', 'wb') as f:
                pickle.dump({'features': arr[start_idx:end_idx], 'metadata': metadata, 'syllables': num_syls}, f)
    
    # Display completion message
    if num_files is None or num_files <= 1:
        messagebox.showinfo("Info", f"Features and metadata saved as {dataset_name}_seg.pkl")
    else:
        messagebox.showinfo("Info", f"Features and metadata saved as {dataset_name}_0_seg.pkl to {dataset_name}_{num_files-1}_seg.pkl")


def remove_pkl_suffix(filename):
    """ Remove the suffix from a filename if it exists. """
    return filename.rsplit('.pkl', 1)[0] if filename.endswith('.pkl') else filename



# utils/audio_utils.py
import numpy as np


def seconds_to_index(seconds, sample_rate):
    """Convert time in seconds to index based on the sample rate."""
    index_size = int(seconds * sample_rate // 1000)
    return index_size


def seconds_to_chunk_index(seconds, chunk_size, sample_rate):
    """Convert time in seconds to chunk index based on the sample rate and chunk size."""
    index_size = (seconds * sample_rate) // chunk_size
    return index_size


def index_to_seconds(index, chunk_size, sample_rate):
    """Convert index to time in seconds based on the chunk size and sample rate."""
    seconds = (index * chunk_size) / sample_rate
    return seconds


def decibel(x, xref=1.0):
    """Convert amplitude values to decibel (dB)."""
    x_copy = np.copy(x)
    x_copy[x_copy < 1e-10] = 1e-10  # Avoid log(0) by setting a minimum threshold
    db = 20.0 * np.log10(x_copy / xref)
    return db
